# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

import logging
import logging.handlers

import os
import errno
#from .system_info import mkdir_p
def mkdir_p(path):
    """http://stackoverflow.com/a/600612/190597 (tzot)"""
    try:
        os.makedirs(path, exist_ok=True)  # Python>3.2
    except TypeError:
        try:
            os.makedirs(path)
        except OSError as exc: # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else: raise

def get_level_from_string(level_string):
    if (level_string == "DEBUG"):
        return logging.DEBUG
    if (level_string == "INFO"):
        return logging.INFO
    if (level_string == "WARNING"):
        return logging.WARNING
    if (level_string == "ERROR"):
        return logging.ERROR
    if (level_string == "CRITICAL"):
        return logging.CRITICAL

def get_app_firewall_payloads_logger():
    from tcell_agent.config import CONFIGURATION
    if not CONFIGURATION.log_appfirewall_events:
        return None

    new_logger = logging.getLogger("app_firewall_payloads_logger")
    formatter = logging.Formatter('%(asctime)s - [%(levelname)s]: %(message)s')

    log_filename = CONFIGURATION.app_firewall_payloads_log_filename

    fh = logging.handlers.RotatingFileHandler(
              log_filename, maxBytes=10*1024*1024, backupCount=5)
    fh.setFormatter(formatter)
    new_logger.addHandler(fh)

    new_logger.setLevel(logging.INFO)

    return new_logger

def get_logger():
    from tcell_agent.config import CONFIGURATION
    logging_options = CONFIGURATION.logging_options
    new_logger = logging.getLogger("tcell_agent")
    formatter = logging.Formatter('%(asctime)s - %(name)s -- [%(levelname)s]: %(message)s')

    if (logging_options is None or logging_options.get("enabled",True) == False):
        return None

    log_filename = CONFIGURATION.log_filename

    fh = logging.handlers.RotatingFileHandler(
              log_filename, maxBytes=10*1024*1024, backupCount=5)
    fh.setFormatter(formatter)

    new_logger.setLevel(get_level_from_string(logging_options.get("level","INFO")))
    new_logger.addHandler(fh)

    return new_logger

try:
    base_logger
    base_app_firewall_logger
except:
    base_logger = get_logger()
    base_app_firewall_logger = get_app_firewall_payloads_logger()



# -*- coding: utf-8 -*-

from tcell_agent.config import TcellAgentConfiguration

import unittest
from ...agent import TCellAgent
import os

class TcellAgentConfigurationTest(unittest.TestCase):
    def test_setting_max_data_ex_db_records_per_request(self):
        full_path = os.path.realpath(__file__)
        data_exposure_config = os.path.join(os.path.dirname(full_path), "data_exposure.config")
        configuration = TcellAgentConfiguration(data_exposure_config)

        self.assertEqual(configuration.max_data_ex_db_records_per_request, 100)



from __future__ import unicode_literals
from __future__ import print_function

import unittest
from ....policies.httptx_policy import HttpTxPolicy

class HttpTxPolicyTest(unittest.TestCase):
    def min_header_test(self):
        policy_json = {"policy_id":"xyzd"}
        policy = HttpTxPolicy()
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.policy_id, "xyzd")
        self.assertEqual(policy.firehose["enabled"], False)

    def new_header_test(self):
        policy_json = {"policy_id":"nyzd", "types":{"firehose":{"enabled":True}}}
        policy = HttpTxPolicy()
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.policy_id, "nyzd")
        self.assertEqual(policy.firehose["enabled"], True)



from tcell_agent.policies.appsensor.sensor import sendEvent
from tcell_agent.sanitize import SanitizeUtils

class LoginSensor(object):
    LOGIN_FAILURE_DP = "lgnFlr"

    def __init__(self, policy_json=None):
        self.enabled = False

        if policy_json is not None:
            self.enabled = policy_json.get("enabled", False)

    def check(self, appsensor_meta, username):
        if not self.enabled:
            return

        if username is not None:
            username = SanitizeUtils.hmac(username)

        sendEvent(appsensor_meta, self.LOGIN_FAILURE_DP, username, None, payload=None)

    def __str__(self):
        return "<%s enabled: %s>" % (type(self).__name__, self.enabled)



from tcell_agent.policies.appsensor import SizeSensor

class RequestSizeSensor(SizeSensor):
    MAX_NORMAL_REQUEST_BYTES = 1024*512
    DP_UNUSUAL_REQUEST_SIZE = "reqsz"

    def __init__(self, policy_json=None):
        super(RequestSizeSensor, self).__init__(
            self.MAX_NORMAL_REQUEST_BYTES,
            self.DP_UNUSUAL_REQUEST_SIZE,
            policy_json
        )



from tcell_agent.policies.appsensor import InjectionSensor

class RetrSensor(InjectionSensor):

    def __init__(self, policy_json=None):
        super(RetrSensor, self).__init__("retr", policy_json)



import json
from .appsensor_rule_set import AppSensorRuleSet

class AppSensorRuleManager(object):
    def __init__(self, filename=None):
        self.rule_info = {}
        if (filename):
            self.load_rules_file(filename)

    def load_rules_file(self, filename):
        self.rule_info.clear()
        with open(filename, 'r') as f:
            rules_from_file = json.loads(f.read())
            rule_types = rules_from_file.get("sensors")
            if (rule_types):
                for rule_type in rule_types:
                    rule_info_json = rule_types[rule_type]

                    rule_set = AppSensorRuleSet()
                    rule_set.set_safe_pattern_from_string(rule_info_json.get("safe_pattern"))

                    for rule_pattern_dict in rule_info_json.get("patterns",[]):
                        rule_set.add_pattern_from_dict(rule_pattern_dict)
                    self.rule_info[rule_type] = rule_set

    def get_ruleset_for(self, rule_type):
        return self.rule_info.get(rule_type)



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from tcell_agent.policies import TCellPolicy

class LoginPolicy(TCellPolicy):
    api_identifier = "login"

    def __init__(self, policy_json=None):
        super(LoginPolicy, self).__init__()
        self.init_variables()
        if policy_json is not None:
            self.loadFromJson(policy_json)

    def init_variables(self):
        self.login_failed_enabled = False
        self.login_success_enabled = False
        self.session_hijacking_metrics = False

    def loadFromJson(self, policy_json):
        if "policy_id" in policy_json:
            self.policy_id = policy_json["policy_id"]
        else:
            raise Exception("Policy Id Not Found")
        self.init_variables()
        policy_data = policy_json.get("data")
        if policy_data:
            options_json = policy_data.get("options")
            if options_json:
                self.login_failed_enabled = options_json.get("login_failed_enabled", False)
                self.login_success_enabled = options_json.get("login_success_enabled", False)
                self.session_hijacking_metrics = options_json.get("session_hijacking_enabled", False)


from __future__ import unicode_literals
from __future__ import print_function
import unittest
from ...sensor_events import DlpEvent

class DlpEventTest(unittest.TestCase):
    def test_dlp_event_create_framework(self):
        dlpe = DlpEvent("routeid", "/rawuri?y=z&d=f", DlpEvent.FOUND_IN_LOG).for_framework(DlpEvent.FRAMEWORK_VARIABLE_SESSION_ID)
        dlpe.post_process()

        self.assertEqual(dlpe["event_type"], "dlp")
        self.assertEqual(dlpe["type"], "framework")
        self.assertEqual(dlpe["variable"], DlpEvent.FRAMEWORK_VARIABLE_SESSION_ID)

    def test_dlp_event_create_request(self):
        dlpe = DlpEvent("routeid", "/rawuri?y=z&d=f", DlpEvent.FOUND_IN_LOG).for_request(DlpEvent.REQUEST_CONTEXT_FORM,"passwd")
        dlpe.post_process()

        self.assertEqual(dlpe["event_type"], "dlp")
        self.assertEqual(dlpe["type"], "request")
        self.assertEqual(dlpe["context"], "form")
        self.assertEqual(dlpe["variable"], "passwd")



# -*- coding: utf-8 -*-

try:
    import django
    from django.conf import settings
    settings.configure()
    settings.DATABASES = {
        'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': 'mydatabase',
        }
    }
    django.setup()
except RuntimeError:
    print("Django already setup")

from django.test import TestCase
from django.test.client import RequestFactory
from django.http import HttpResponse
from django.middleware.common import CommonMiddleware
from django.conf import settings

from mock import Mock

from .....instrumentation.django.middleware.globalrequestmiddleware import GlobalRequestMiddleware
from .....instrumentation.django.middleware.afterauthmiddleware import AfterAuthMiddleware
from .....instrumentation.django.middleware.body_filter_middleware import BodyFilterMiddleware
from .....config import CONFIGURATION
from .....agent import TCellAgent


try:
    settings.configure()
except:
    pass

settings.DEFAULT_CHARSET = 'utf-8'


class BodyFilterMiddlewareTest(TestCase):
    _multiprocess_can_split_ = False

    def setUp(self):
        settings.ALLOWED_HOSTS = ['testserver', 'test.tcell.io']

        self.grm = GlobalRequestMiddleware()
        self.aam = AfterAuthMiddleware()
        self.cm = CommonMiddleware()
        self.bfm = BodyFilterMiddleware()

        rf = RequestFactory()
        self.request = rf.get('http://test.tcell.io/hello/')


        self.request.session = Mock()
        self.request.session.session_key = "101012301200123"

        CONFIGURATION.version = 1
        CONFIGURATION.api_key = "Test_-ApiKey=="
        CONFIGURATION.app_id = "TestAppId-AppId"
        CONFIGURATION.host_identifier = "TestHostIdentifier"
        CONFIGURATION.uuid = "test-uuid-test-uuid"
        CONFIGURATION.fetch_policies_from_tcell = False

    def test_body_inject(self):
        def addEvent(self, event):
          event.post_process()

        TCellAgent.addEvent = addEvent
        TCellAgent.tCell_agent = TCellAgent()

        policy_json = {
            "csp-headers":{
                "policy_id":"nyzd",
                "data":{
                    "options":{
                        "js_agent_api_key":"000-000-1-2323"
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        self.grm.process_request(self.request)
        self.aam.process_request(self.request)
        self.cm.process_request(self.request)
        self.bfm.process_request(self.request)
        response = HttpResponse("<html>\n<head>Title</head><body>hello world</body><html>", content_type="text/html")
        self.bfm.process_response(self.request, response)
        self.cm.process_response(self.request, response)
        self.aam.process_response(self.request, response)
        self.grm.process_response(self.request, response)
        expected = b'<html>\n<head><script src="https://api.tcell.io/tcellagent.min.js" tcellappid="TestAppId-AppId" tcellapikey="000-000-1-2323"></script>Title</head><body>hello world</body><html>'
        self.assertEqual(response.content, expected)

    def test_non_html_body_inject(self):
        def addEvent(self, event):
          event.post_process()

        TCellAgent.addEvent = addEvent
        TCellAgent.tCell_agent = TCellAgent()

        policy_json = {
            "csp-headers":{
                "policy_id":"nyzd",
                "data":{
                    "options":{
                        "js_agent_api_key":"000-000-1-2323"
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        self.grm.process_request(self.request)
        self.aam.process_request(self.request)
        self.cm.process_request(self.request)
        self.bfm.process_request(self.request)
        resp = b'\x08\x01\x10\x01(\xc1\xca\x03JN\n\tajaxtoken\x1282cfc663150b955a7b8aad324f33364d5965f1a7397a545df76dea8bf\x18\x00 \x80\xc4\xff\x0e'
        response = HttpResponse(resp, content_type="text/html")
        self.bfm.process_response(self.request, response)
        self.cm.process_response(self.request, response)
        self.aam.process_response(self.request, response)
        self.grm.process_response(self.request, response)
        expected = b"""\x08\x01\x10\x01(\xc1\xca\x03JN\n\tajaxtoken\x1282cfc663150b955a7b8aad324f33364d5965f1a7397a545df76dea8bf\x18\x00 \x80\xc4\xff\x0e"""
        self.assertEqual(response.content, expected)

    def test_utf8_html_body_inject(self):
        def addEvent(self, event):
          event.post_process()

        TCellAgent.addEvent = addEvent
        TCellAgent.tCell_agent = TCellAgent()

        policy_json = {
            "csp-headers":{
                "policy_id":"nyzd",
                "data":{
                    "options":{
                        "js_agent_api_key":"000-000-1-2323"
                    }
                }
            }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        self.grm.process_request(self.request)
        self.aam.process_request(self.request)
        self.cm.process_request(self.request)
        self.bfm.process_request(self.request)
        response = HttpResponse(u"<html>\n<head>Müller</head><body>holá mundo</body><html>", content_type="text/html")
        self.bfm.process_response(self.request, response)
        self.cm.process_response(self.request, response)
        self.aam.process_response(self.request, response)
        self.grm.process_response(self.request, response)
        expected = b'<html>\n<head><script src="https://api.tcell.io/tcellagent.min.js" tcellappid="TestAppId-AppId" tcellapikey="000-000-1-2323"></script>M\xc3\xbcller</head><body>hol\xc3\xa1 mundo</body><html>'
        self.assertEqual(response.content, expected)



from __future__ import unicode_literals
from __future__ import print_function

import unittest

from mock import patch
from tcell_agent.appsensor.meta import AppSensorMeta
from tcell_agent.policies.appsensor import UserAgentSensor

class UserAgentSensorTest(unittest.TestCase):

    def create_default_sensor_test(self):
        sensor = UserAgentSensor()
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.empty_enabled, False)

    def create_enabled_sensor_test(self):
        sensor = UserAgentSensor({"enabled": True})
        self.assertEqual(sensor.enabled, True)

    def create_empty_enabled_sensor_test(self):
        sensor = UserAgentSensor({"empty_enabled": True})
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.empty_enabled, True)

    def with_disabled_sensor_check_test(self):
        sensor = UserAgentSensor({"enabled": False})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.user_agent_sensor.sendEvent') as patched_send_event:
            sensor.check(meta)
            self.assertFalse(patched_send_event.called)

    def with_enabled_sensor_and_user_agent_is_present_check_test(self):
        sensor = UserAgentSensor({"enabled": True, "empty_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        meta.request_data(1024, {}, {}, {}, "", "user agent", {})
        with patch('tcell_agent.policies.appsensor.user_agent_sensor.sendEvent') as patched_send_event:
            sensor.check(meta)
            self.assertFalse(patched_send_event.called)

    def with_enabled_sensor_and_user_agent_is_none_check_test(self):
        sensor = UserAgentSensor({"enabled": True, "empty_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        meta.request_data(1024, {}, {}, {}, "", None, {})
        with patch('tcell_agent.policies.appsensor.user_agent_sensor.sendEvent') as patched_send_event:
            sensor.check(meta)
            patched_send_event.assert_called_once_with(meta, UserAgentSensor.DP_CODE, None, None)

    def with_enabled_sensor_and_user_agent_is_empty_string_check_test(self):
        sensor = UserAgentSensor({"enabled": True, "empty_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        meta.request_data(1024, {}, {}, {}, "", "", {})
        with patch('tcell_agent.policies.appsensor.user_agent_sensor.sendEvent') as patched_send_event:
            sensor.check(meta)
            patched_send_event.assert_called_once_with(meta, UserAgentSensor.DP_CODE, None, None)

    def with_enabled_sensor_and_user_agent_is_blank_string_check_test(self):
        sensor = UserAgentSensor({"enabled": True, "empty_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        meta.request_data(1024, {}, {}, {}, "", "   \t \n", {})
        with patch('tcell_agent.policies.appsensor.user_agent_sensor.sendEvent') as patched_send_event:
            sensor.check(meta)
            patched_send_event.assert_called_once_with(meta, UserAgentSensor.DP_CODE, None, None)

    def with_enabled_sensor_and_user_agent_is_empty_string_and_matching_exluded_route_check_test(self):
        sensor = UserAgentSensor({"enabled": True, "empty_enabled": True, "exclude_routes": ["23947"]})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        meta.request_data(1024, {}, {}, {}, "", "", {})
        with patch('tcell_agent.policies.appsensor.user_agent_sensor.sendEvent') as patched_send_event:
            sensor.check(meta)
            self.assertFalse(patched_send_event.called)

    def with_enabled_sensor_and_user_agent_is_empty_string_and_nonmatching_exluded_route_check_test(self):
        sensor = UserAgentSensor({"enabled": True, "empty_enabled": True, "exclude_routes": ["nonmatching"]})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        meta.request_data(1024, {}, {}, {}, "", "", {})
        with patch('tcell_agent.policies.appsensor.user_agent_sensor.sendEvent') as patched_send_event:
            sensor.check(meta)
            patched_send_event.assert_called_once_with(meta, UserAgentSensor.DP_CODE, None, None)



from __future__ import unicode_literals
from __future__ import print_function
import unittest
from ...sensor_events import ServerAgentPackagesEvent
#from ... import tcell_agent
import json
from ...api import SetEncoder

class MaxLengthTest(unittest.TestCase):
    def test_package_event_create(self):
        sape = ServerAgentPackagesEvent()
        sape.add_package("test_package", "t"*400)
        x = json.loads(json.dumps(sape, cls=SetEncoder))
        self.assertEqual(x["packages"][0]["v"],"t"*256)




# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

from tcell_agent.agent import TCellAgent, PolicyTypes
from tcell_agent.appsensor.manager import app_sensor_manager
from tcell_agent.appsensor.meta import AppSensorMeta
from tcell_agent.instrumentation import safe_wrap_function

import json

def django_meta(request):
  meta = AppSensorMeta(
      request._tcell_context.remote_addr ,
      request.META.get("REQUEST_METHOD"),
      request.build_absolute_uri(),
      request._tcell_context.route_id,
      session_id=request._tcell_context.session_id,
      user_id=request._tcell_context.user_id
  )
  return meta

def appsensor_login_failed(request, username):
    appsensor_policy = TCellAgent.get_policy(PolicyTypes.APPSENSOR)
    if appsensor_policy is None:
        return
    meta = django_meta(request)
    appsensor_policy.check_login_failure(meta, username)

def django_request_appsensor(request):
    appsensor_policy = TCellAgent.get_policy(PolicyTypes.APPSENSOR)
    if appsensor_policy is None:
        return
    meta = django_meta(request)
    request_len = None

    try:
        request_len = int(request.META.get("CONTENT_LENGTH",0))
    except:
        pass

    post_dict = {}
    request_json_body = None

    try:
        if (request_len is not None and request_len > 0):
            post_dict = request.POST
            content_type = request.META.get("CONTENT_TYPE","")
            if content_type.startswith("multipart/form-data") == False:
                # Can't just say post as it may be PUT or maybe something else
                # We're going to make sure some crazy client didn't submit json by mistake
                request_body = request.body
                if (request_len < 2000000 and len(request_body) < 2000000):
                    if content_type.startswith("application/json"):
                        request_json_body = request_body
                    else:
                        if isinstance(request_body, bytes):
                            request_body = request_body.decode('utf-8')
                        if request_body[0] == '{' or request_body[0] == '[':
                            try:
                                json.loads(request_body)
                                post_dict = {}
                                request_json_body = request_body
                            except ValueError as ve:
                                pass
    except Exception as e:
        # Log this?
        pass

    meta.request_data(
        request_len,
        request.GET,
        post_dict,
        request.COOKIES,
        request_json_body,
        request.META.get("HTTP_USER_AGENT"),
        request.FILES
    )
    app_sensor_manager.send_appsensor_data(meta)

def django_response_appsensor(django_response, request, response):
    appsensor_policy = TCellAgent.get_policy(PolicyTypes.APPSENSOR)
    if appsensor_policy is None:
        return
    meta = django_meta(request)
    response_content_len = 0
    try:
        if isinstance(response, django_response):
            response_content_len = len(response.content)
    except:
        pass
    response_code = response.status_code
    meta.response_data(response_content_len, response_code)
    app_sensor_manager.send_appsensor_data(meta)

def django_path_parameters_appsensor(request, path_dict):
    appsensor_policy = TCellAgent.get_policy(PolicyTypes.APPSENSOR)
    if appsensor_policy is None:
        return
    meta = django_meta(request)
    meta.path_parameters_data(path_dict)
    app_sensor_manager.send_appsensor_data(meta)



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

import tcell_agent
from tcell_agent.agent import TCellAgent, PolicyTypes
import os
from tcell_agent.sensor_events import HttpTxSensorEvent, FingerprintSensorEvent, LoginSuccessfulSensorEvent, LoginFailureSensorEvent
from tcell_agent.sensor_events import RedirectSensorEvent
from tcell_agent.sensor_events import ServerAgentDetailsEvent
from .settings import send_django_setting_events
import uuid


from tcell_agent.sanitize import SanitizeUtils

from future.backports.urllib.parse import urlsplit
from future.backports.urllib.parse import urlunsplit
from future.backports.urllib.parse import parse_qs
_started = False

from .routes import make_route_table

from tcell_agent.instrumentation.django.middleware.globalrequestmiddleware import GlobalRequestMiddleware
from tcell_agent.instrumentation.django.middleware import globalrequestmiddleware
from tcell_agent.instrumentation.django.middleware.csrf_exception_middleware import instrument_csrf_view_middleware
from tcell_agent.instrumentation.django.database_error_wrapper import instrument_database_error_wrapper

import logging
import tcell_agent.tcell_logger
from tcell_agent.instrumentation.django.utils import django15or16

LOGGER = logging.getLogger('tcell_agent').getChild(__name__)

try:
    if django15or16:
        dlp_success = False
    else:
        from tcell_agent.instrumentation.django.dlp import dlp_instrumentation
        dlp_success = True
except Exception as e:
    LOGGER.debug(e)
    dlp_success = False



def _instrument():
    from django.core.handlers.base import BaseHandler
    old_load_middleware = BaseHandler.load_middleware
    def load_middleware(*args, **kwargs):
        LOGGER.info("Adding middleware")
        _insertMiddleware(
            'tcell_agent.instrumentation.django.middleware.body_filter_middleware.BodyFilterMiddleware')
        _insertMiddleware(
            'tcell_agent.instrumentation.django.middleware.afterauthmiddleware.AfterAuthMiddleware')
        _insertMiddleware(
            'tcell_agent.instrumentation.django.middleware.tcelllastmiddleware.TCellLastMiddleware')
        _insertMiddleware(
          'tcell_agent.instrumentation.django.middleware.tcell_data_exposure_middleware.TCellDataExposureMiddleware',
          atIdx=0)
        _insertMiddleware(
          'tcell_agent.instrumentation.django.middleware.globalrequestmiddleware.GlobalRequestMiddleware',
          atIdx=0)
        _insertMiddleware(
             'tcell_agent.instrumentation.django.middleware.timermiddleware.TimerMiddleware')

        if dlp_success == True:
            dlp_instrumentation()

        if _is_csrf_middleware_enabled():
            instrument_csrf_view_middleware()

        instrument_database_error_wrapper()

        import tcell_agent.instrumentation.django.contrib_auth
        return old_load_middleware(*args, **kwargs)
    BaseHandler.load_middleware =  load_middleware

    from django.middleware.common import CommonMiddleware
    old_func = CommonMiddleware.__init__
    def init(self):
        if (tcell_agent.instrumentation.django._started == False):
            LOGGER.info("Starting agent")

            tcell_agent.instrumentation.django._started = True
            TCellAgent.get_agent().ensure_polling_thread_running()

            sade = ServerAgentDetailsEvent(framework="Django",framework_version=django.get_version())
            TCellAgent.send(sade)
            make_route_table()
            send_django_setting_events()
        return old_func(self)
    CommonMiddleware.__init__ = init

def _insertMiddleware(middlewareClassString, after=None, before=None, atIdx=None):
    from django.conf import settings
    middleware_list = list(settings.MIDDLEWARE_CLASSES)
    if after:
        idx = middleware_list.index(after)+1 if after in middleware_list else len(middleware_list)
    elif before:
        idx = middleware_list.index(before) if before in middleware_list else 0
    elif atIdx is not None:
        idx = atIdx
    else:
        idx = len(middleware_list)
    middleware_list.insert(idx, middlewareClassString)
    settings.MIDDLEWARE_CLASSES = tuple(middleware_list)

def _is_csrf_middleware_enabled():
    from django.conf import settings
    return "django.middleware.csrf.CsrfViewMiddleware" in list(settings.MIDDLEWARE_CLASSES)

try:
    import django
    if TCellAgent.get_agent():
        _instrument()

except Exception as e:
    LOGGER.debug("Could not instrument django")
    LOGGER.debug(e)




# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

""" Agent Module handles communication and instrumentation, this
is the main class.
"""
from __future__ import unicode_literals
from __future__ import print_function

from ..config import CONFIGURATION
import os

class DiscoveryCache:
    def current_cache_json(self, discovery_type):
        cache_filename = os.path.join(CONFIGURATION.tcell_folder, CONFIGURATION.cache_folder, "discovery_cache", ".", discovery_cache)
        if os.path.isfile(cache_filename) == True:
            cache_json = {}
        try:
            with open(cache_filename) as data_file:
                json_data = data_file.read()
                cache_json = json.loads(json_data)
        except Exception as e:
            LOGGER.error("Error loading configuration file")
            LOGGER.exception(e)
        return cache_json

    def update_cache(self, discovery_type, discovery_table_json):
        cache_filename = os.path.join(CONFIGURATION.tcell_folder, CONFIGURATION.cache_folder, "discovery_cache", ".", discovery_cache)
        try:
            with open(self.cache_filename, "w") as data_file:
                json.dump(discovery_table_json, data_file, indent=2)
            data_file.close()
        except Exception as e:
            LOGGER.error("Error writing configuration file")
            LOGGER.exception(e)
        return




# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function
from .configuration import TcellAgentConfiguration
try:
    CONFIGURATION
except NameError:
    CONFIGURATION = TcellAgentConfiguration()



from .base_policy import TCellPolicy
from .clickjacking_policy import ClickjackingPolicy
from .content_security_policy import ContentSecurityPolicy
from .http_redirect_policy import HttpRedirectPolicy
from .httptx_policy import HttpTxPolicy
from .secure_headers_policy import SecureHeadersPolicy
from .dataloss_policy import DataLossPolicy
from .appsensor_policy import AppSensorPolicy
from .login_policy import LoginPolicy
from .honeytoken_policy import HoneytokenPolicy
from .patches_policy import PatchesPolicy



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function


import re

from future.backports.urllib.parse import urlsplit

from tcell_agent.policies import TCellPolicy
from tcell_agent.sensor_events.http_redirect import RedirectSensorEvent


def get_location_host(url):
    scheme, netloc, path, query_string, fragment = urlsplit(url)
    return netloc
    #return urlunsplit(("", netloc, "", "", ""))

class HttpRedirectPolicy(TCellPolicy):
    def __init__(self, policy_json=None):
        super(HttpRedirectPolicy, self).__init__()
        self.enabled = False
        self.block = False
        self.whitelist = []
        if policy_json is not None:
            self.loadFromJson(policy_json)

    def loadFromJson(self, policy_json):
        if "policy_id" in policy_json:
            self.policy_id = policy_json["policy_id"]
        else:
            raise Exception("Policy Id Not Found")

        policy_data_json = policy_json.get("data")
        if policy_data_json:
            self.enabled = policy_data_json.get("enabled", False)
            self.block = policy_data_json.get("block", False)
            whitelist = policy_data_json.get("whitelist", [])
            self.whitelist = [self.wildcard_re(item) for item in whitelist]

    def wildcard_re(self, item):
        return re.compile("^" + re.escape(item).replace("\*",".*") + "$")

    def process_location(self,
                         remote_addr,
                         method,
                         from_domain,
                         from_full_path,
                         status_code,
                         redirect_url,
                         user_id=None,
                         session_id=None,
                         route_id=None):

        from tcell_agent.agent import TCellAgent

        location_host = get_location_host(redirect_url)

        if location_host == "":
            return redirect_url

        if location_host == from_domain:
            return redirect_url

        for item in self.whitelist:
            if item.match(location_host):
                return redirect_url

        event = RedirectSensorEvent(
            remote_addr,
            method,
            from_domain,
            from_full_path,
            status_code,
            location_host,
            user_id,
            session_id,
            route_id)

        TCellAgent.tCell_agent.addEvent(event)

        if self.block == True:
            return "/"

        return redirect_url



from tcell_agent.appsensor.rules import rule_manager
from tcell_agent.policies.appsensor import InjectionSensor

class NullbyteSensor(InjectionSensor):

    def __init__(self, policy_json=None):
        super(NullbyteSensor, self).__init__("null", policy_json)

    def get_ruleset(self):
        return rule_manager.get_ruleset_for("nullbyte")



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

from tcell_agent.agent import TCellAgent, PolicyTypes
import os
from tcell_agent.sensor_events.httptx import HttpTxSensorEvent, FingerprintSensorEvent, LoginSuccessfulSensorEvent, LoginFailureSensorEvent
from tcell_agent.sensor_events.http_redirect import RedirectSensorEvent
import uuid
import re

from tcell_agent.sanitize import SanitizeUtils
from tcell_agent.instrumentation import handle_exception, safe_wrap_function

from future.backports.urllib.parse import urlsplit
from future.backports.urllib.parse import urlunsplit
from future.backports.urllib.parse import parse_qs

from tcell_agent.instrumentation.context import TCellInstrumentationContext

from django.http import HttpResponse
from django.http import HttpResponseForbidden
import time
import threading
import logging
logger = logging.getLogger('tcell_agent').getChild(__name__)

from django.conf import settings
from django.core.urlresolvers import get_resolver, set_urlconf
from tcell_agent.instrumentation.django.routes import get_route_table
from tcell_agent.config import CONFIGURATION


def get_location_host(url):
    scheme, netloc, path, query_string, fragment = urlsplit(url)
    return urlunsplit((scheme, netloc, "", "", ""))

class GlobalRequestMiddleware(object):
    _threadmap = {}

    @classmethod
    def get_current_request(cls):
      try:
        return cls._threadmap[threading.current_thread().ident]
      except Exception as e:
        pass

    def process_request(self, request):
        request._tcell_signals = {}

        request._tcell_context = TCellInstrumentationContext()


        def block_ip():
            patches_policy = TCellAgent.get_policy(PolicyTypes.PATCHES)
            if patches_policy:
                return patches_policy.is_ip_blocked(request._tcell_context.remote_addr)
            return False

        def assign_route_id():
            urlconf = settings.ROOT_URLCONF
            set_urlconf(urlconf)
            resolver = get_resolver(urlconf)
            try:
                resolver_match = resolver.resolve(request.path_info)
            except:
                return                
            if resolver_match:
                route_table = get_route_table()
                request._tcell_context.route_id = route_table.get(resolver_match.func,{}).get("route_id")

        def better_ip_address():
            if CONFIGURATION.reverse_proxy != True:
                return request.META.get('REMOTE_ADDR','1.1.1.1')
            else:
                try:
                    reverse_proxy_header = CONFIGURATION.reverse_proxy_ip_address_header
                    if reverse_proxy_header is None:
                        reverse_proxy_header = "HTTP_X_FORWARDED_FOR" # This is the standard
                    else:
                        reverse_proxy_header = "HTTP_" + reverse_proxy_header.upper().replace('-','_')
                    x_forwarded_for = request.META.get(reverse_proxy_header,request.META.get('REMOTE_ADDR','1.1.1.1'))
                    if x_forwarded_for and x_forwarded_for != '':
                        ip = x_forwarded_for.split(',')[0].strip()
                    else:
                        ip = request.META.get('REMOTE_ADDR','1.1.1.1')
                    return ip
                except:
                    return request.META.get('REMOTE_ADDR','1.1.1.1')

        def initialize_request():
            request._tcell_context.transaction_id = str(uuid.uuid4())
            request._tcell_context.user_agent = request.META.get("HTTP_USER_AGENT")
            request._tcell_context.remote_addr = better_ip_address()


        def set_fullpath():
            request._tcell_context.fullpath = request.get_full_path()

        safe_wrap_function("Setting Transaction Id", initialize_request)

        if safe_wrap_function("Checking ip blocking", block_ip):
            return HttpResponseForbidden()

        safe_wrap_function("Setting Request FullPath", set_fullpath)
        safe_wrap_function("Assiging route ID to request (global)", assign_route_id)

        self._threadmap[threading.current_thread().ident] = request

    def process_exception(self, request, exception):
        try:
            del self._threadmap[threading.current_thread().ident]
        except KeyError:
            pass

    def process_response(self, request, response):
        safe_wrap_function("Check Location Header", self.checkLocationRedirect, request, response)
        safe_wrap_function("Insert CSP Headers", self.addCSP, request, response)
        safe_wrap_function("Insert Clickjacking Headers", self.addClickjacking, request, response)
        safe_wrap_function("Insert Secure Headers", self.addSecureHeaders, request, response)
        safe_wrap_function("Running Filters", self.run_filters, request, response)
        try:
            del self._threadmap[threading.current_thread().ident]
        except KeyError:
            pass        
        return response

    def run_filters(self, request, response):
        if(isinstance(response, HttpResponse)):
            response.content = request._tcell_context.filter_body(response.content)

    def addSecureHeaders(self, request, response):
        if TCellAgent.get_policy(PolicyTypes.SECURE_HEADERS):
            secure_headers = TCellAgent.get_policy(PolicyTypes.SECURE_HEADERS).headers()
            if secure_headers:
                for secure_header in secure_headers:
                    response[secure_header[0]] = secure_header[1]

    def addCSP(self, request, response):
        if(isinstance(response, HttpResponse) and response.has_header("Content-Type") ):
            if response["Content-Type"] and response["Content-Type"].startswith("text/html"):
                if TCellAgent.get_policy(PolicyTypes.CSP):
                    csp_headers = TCellAgent.get_policy(PolicyTypes.CSP).headers(
                        request._tcell_context.transaction_id,
                        request._tcell_context.route_id,
                        request._tcell_context.session_id,
                        request._tcell_context.user_id)
                    if csp_headers:
                        for csp_header in csp_headers:
                            header_value = csp_header[1]
                            response[csp_header[0]] = header_value

    def addClickjacking(self, request, response):
        if TCellAgent.get_policy(PolicyTypes.CLICKJACKING):
            csp_headers = TCellAgent.get_policy(PolicyTypes.CLICKJACKING).headers(
                request._tcell_context.transaction_id,
                request._tcell_context.route_id,
                request._tcell_context.session_id,
                request._tcell_context.user_id)
            if csp_headers:
                for csp_header in csp_headers:
                    header_value = csp_header[1]
                    if (response.has_header(csp_header[0])):
                        response[csp_header[0]] =  response[csp_header[0]] + ", " + header_value
                    else:
                        response[csp_header[0]] = header_value


    def checkLocationRedirect(self, request, response):
        redirect_policy = TCellAgent.get_policy(PolicyTypes.HTTP_REDIRECT)

        if redirect_policy and response.get("location") and redirect_policy.enabled:
            response["location"] = redirect_policy.process_location(
                request._tcell_context.remote_addr,
                request.method,
                request.get_host(),
                request.get_full_path(),
                response.status_code,
                response.get("location"),
                user_id=request._tcell_context.user_id,
                session_id=request._tcell_context.session_id,
                route_id=request._tcell_context.route_id)

    def handleFingerprint(self, request, response):
        #if TCellAgent.tCell_agent.httptx_policy.fingerprint.get("enabled") and \
        event = FingerprintSensorEvent(request, request._tcell_context.session_id)
        TCellAgent.tCell_agent.addEvent(event)


    def handleLoginSuccessfulSignal(self, request, response):
        if TCellAgent.get_policy(PolicyTypes.HTTP_TX).auth_framework.get("enabled") and \
        request._tcell_sensors_httpx.get("login_attempt"):
            event = LoginSuccessfulSensorEvent(request, response)
            TCellAgent.send(event)

    def handleLoginFailureSignal(self, request, response):
        if TCellAgent.get_policy(PolicyTypes.HTTP_TX).auth_framework.get("enabled") and \
        request._tcell_sensors_httpx.get("login_attempt"):
            event = LoginFailureSensorEvent(request, response)
            TCellAgent.send(event)
            #elif TCellAgent.tCell_agent.httptx_policy.firehose.get("enabled"):
            #    content_type = response.get("content-type")
            #    if (content_type and content_type.lower().startswith("text/html")):
            #        logger.debug("firehose event")
            #        event = HttpTxSensorEvent(request, response)
            #        TCellAgent.tCell_agent.addEvent(event)



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

""" Agent Module handles communication and instrumentation, this
is the main class.
"""

from __future__ import unicode_literals
from __future__ import print_function

import os, fcntl
import time
import sys
import time
try:
    import cPickle as pickle
except:
    import pickle
import struct

import threading
import traceback
                
import logging
import tcell_agent.tcell_logger
LOGGER = logging.getLogger('tcell_agent').getChild(__name__)

class ForkPipeManager(object):
    def __init__(self, callback=None):
        self.callback = callback
        self.readp, self.writep = os.pipe() 
        self.wr = os.fdopen(self.writep, 'wb', 0)
        self.pipe_running = False
        self.polling_thread = threading.Thread(target=self.run, args=())
        self.polling_thread.daemon = True # Daemonize thread
        self.polling_thread.start()

    def run(self):
        try:
            r = os.fdopen(self.readp, 'rb')
            self.pipe_running = True
            LOGGER.debug("Parent Thread Starting Listen loop {pid}".format(pid=os.getpid()))
            while(True):
              try:
                LOGGER.debug("Parent Thread Waiting to read object {pid}".format(pid=os.getpid()))
                pickled_obj_byte_len = r.read(4)
                length = struct.unpack(">I", pickled_obj_byte_len)[0]
                pickled_obj = r.read(length)
                value = pickle.loads(pickled_obj)
                if self.callback:
                    self.callback(value)
              except EOFError as eofe:
                LOGGER.error("Parent Thread eof exception in pipe read")
                LOGGER.debug(eofe)
                time.sleep(1.0)
              except Exception as eee:
                LOGGER.error("Parent Thread other exception in pipe read")
                LOGGER.debug(eee)
                time.sleep(1.0)
        except Exception as e:
            LOGGER.error("Parent Thread general exception in pipe read")
            LOGGER.debug(e)
        finally:
            self.pipe_running = False

    def send_to_parent(self, obj):
        try:
            LOGGER.debug("Fork sending to parent {pid} {status}".format(pid=os.getpid(), status=self.pipe_running))
            if (self.pipe_running == True):
                pickled_obj = pickle.dumps(obj)
                #if sys.version_info >= (3, 0):
                #    pickled_obj = unicode(pickled_obj)
                pickled_obj_byte_len = struct.pack(">I", len(pickled_obj))
                self.wr.write(pickled_obj_byte_len + pickled_obj)
            else:
                LOGGER.debug("POSSIBLE ISSUE: fork not running.")
                return False
        except Exception as e:
            LOGGER.debug("exception sending to parent pipe")
            LOGGER.debug(e)
            return False
        return True


# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function
import json

class SensorEvent(dict):
    def __init__(self, event_type, flush_right_away=False, ensure_delivery=False, send_event=True, queue_wait=False):
        self.queue_wait = queue_wait
        self.send_event = send_event
        self.flush_right_away = flush_right_away
        self.ensure_delivery = ensure_delivery
        self["event_type"] = event_type

    def get_debug_data(self):
        return "event id: {id} type: {event_type}".format(id=hex(id(self)), event_type=self.get("event_type","<unknown>"))

    def post_process(self):
        # implement here what the background thread should do
        # such as sanitization
        pass

from .app_config_settings import AppConfigSettings
from .http_redirect import RedirectSensorEvent
from .httptx import HttpTxSensorEvent
from .httptx import LoginFailureSensorEvent
from .httptx import LoginSuccessfulSensorEvent
from .httptx import FingerprintSensorEvent
from .server_agent_packages import ServerAgentPackagesEvent
from .server_agent_details import ServerAgentDetailsEvent
from .appsensor import AppSensorEvent
from .app_route import AppRouteSensorEvent
from .dlp import DlpEvent
from .login import LoginEvent
from .metrics import MetricsEvent
from .discovery import DiscoveryEvent
from .honeytoken import HoneytokenSensorEvent


# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function
from . import SensorEvent
from tcell_agent.sanitize import SanitizeUtils
from future.backports.urllib.parse import urlsplit

class ServerAgentDetailsEvent(SensorEvent):
    def __init__(self, 
        user=None, group=None, 
        language=None, language_version=None,
        framework=None, framework_version=None):
        super(ServerAgentDetailsEvent, self).__init__("server_agent_details")
        if user:
            self["user"] = user
        if group:
            self["group"] = group
        if language:
            self["language"] = language
        if language_version:
            self["language_version"] = language_version
        if framework:
            self["app_framework"] = framework
        if framework_version:
            self["app_framework_version"] = framework_version



# coding=utf-8

import unittest

from tcell_agent.appsensor import params

class ParamsTest(unittest.TestCase):

    def one_flatten_params_test(self):
        result = params.flatten_clean({
            "action": "index"
        })

        self.assertEqual(
            result,
            {("action",): "index"}
        )

    def two_flatten_params_test(self):
        result = params.flatten_clean({
            u"müller": "sadness"
        })

        self.assertEqual(
            result,
            {(u"müller",): "sadness"}
        )

    def three_flatten_params_test(self):
        result = params.flatten_clean({
            u"müller".encode("utf-8"): "sadnessdos"
        })

        self.assertEqual(
            result,
            {(b'm\xc3\xbcller',): "sadnessdos"}
        )

    def four_flatten_params_test(self):
        result = params.flatten_clean({
            "utf8-char": u"Müller",
        })

        self.assertEqual(
            result,
            {("utf8-char",): u"Müller"}
        )

    def five_flatten_params_test(self):
        result = params.flatten_clean({
            "waitlist_entries": {"email": "emailone", "preferences": {"email": "emaildos"}}
        })

        self.assertEqual(
            result,
            {
                ("waitlist_entries", "email",): "emailone",
                ("waitlist_entries", "preferences", "email",): "emaildos"
            }
        )

    def six_flatten_params_test(self):
        result = params.flatten_clean({
            "email_preferences": ["daily_digest", "reminders", u"Müller".encode('utf8')],
        })

        self.assertEqual(
            result,
            {
                ('0', "email_preferences",): "daily_digest",
                ('1', "email_preferences",): "reminders",
                ('2', 'email_preferences'): u'M\xfcller',
            }
        )

    def seven_flatten_params_test(self):
        result = params.flatten_clean({
            "users": [
                {"email": "one@email.com"},
                {"email": "dos@email.com"},
            ]
        })

        self.assertEqual(
            result,
            {
                ('0', "users", "email",): "one@email.com",
                ('1', "users", "email",): "dos@email.com"
            }
        )



# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function
from . import SensorEvent

class MetricsEvent(SensorEvent):
    def __init__(self):
        super(MetricsEvent, self).__init__("metrics")
    def set_rct(self, route_count_table):
        self["rct"] = route_count_table
    def set_session(self, session_info):
        if session_info:
            self["sessions"] = session_info


# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

""" Agent Module handles communication and instrumentation, this
is the main class.
"""

from __future__ import unicode_literals
from __future__ import print_function

from future.utils import listvalues

import threading
import time
import json
from queue import Queue, Empty, Full
import os
import tcell_agent
from collections import defaultdict
import traceback
import random

from tcell_agent.config import CONFIGURATION
from tcell_agent.version import VERSION

import logging
import tcell_agent.tcell_logger
LOGGER = logging.getLogger('tcell_agent').getChild(__name__)

from tcell_agent.api import TCellAPI, TCellAPIException
from tcell_agent.policies import ContentSecurityPolicy
from tcell_agent.policies import HttpTxPolicy
from tcell_agent.policies import HttpRedirectPolicy
from tcell_agent.policies import SecureHeadersPolicy
from tcell_agent.policies import ClickjackingPolicy
from tcell_agent.policies import DataLossPolicy
from tcell_agent.policies import AppSensorPolicy
from tcell_agent.policies import LoginPolicy
from tcell_agent.policies import HoneytokenPolicy
from tcell_agent.policies import PatchesPolicy

from tcell_agent.sensor_events import ServerAgentPackagesEvent
from tcell_agent.sensor_events import ServerAgentDetailsEvent
from tcell_agent.sensor_events import MetricsEvent
from tcell_agent.sensor_events import DiscoveryEvent
from tcell_agent.sensor_events import AppRouteSensorEvent
from tcell_agent.sensor_events.flush_dummy import FlushDummyEvent

from tcell_agent.routes import RouteTable

from tcell_agent import system_info
from tcell_agent.policy_cache import TCellPolicyCache

from .fork_manager import ForkPipeManager
from .metrics import AgentMetrics

lock = threading.Lock()
agentLock = threading.Lock()
pollingLock = threading.Lock()
forkThreadsLock = threading.Lock()
eventHandlerThreadsLock = threading.Lock()

class PolicyTypes(object):
    CSP = "csp-headers"
    SECURE_HEADERS = "secure-headers"
    HTTP_TX = "http-tx"
    HTTP_REDIRECT = "http-redirect"
    CLICKJACKING = "clickjacking"
    DATALOSS = DataLossPolicy.api_identifier
    APPSENSOR = AppSensorPolicy.api_identifier
    LOGIN = LoginPolicy.api_identifier
    HONEYTOKEN = HoneytokenPolicy.api_identifier
    PATCHES = PatchesPolicy.api_identifier

class TCellAgent(object):
    """Agent Class for holding and managing threads."""
    policy_classes = {
            PolicyTypes.CSP: ContentSecurityPolicy,
            PolicyTypes.SECURE_HEADERS: SecureHeadersPolicy,
            PolicyTypes.HTTP_TX: HttpTxPolicy,
            PolicyTypes.HTTP_REDIRECT: HttpRedirectPolicy,
            PolicyTypes.CLICKJACKING: ClickjackingPolicy,
            PolicyTypes.DATALOSS: DataLossPolicy,
            PolicyTypes.APPSENSOR: AppSensorPolicy,
            PolicyTypes.LOGIN: LoginPolicy,
            PolicyTypes.HONEYTOKEN: HoneytokenPolicy,
            PolicyTypes.PATCHES: PatchesPolicy
            }
    tCell_agent = None
    def __init__(self):
        """Initialize policies and threads"""
        LOGGER.info("Initializing agent")
        try:
            self.parent_pid = os.getpid()
        except Exception as ex:
            LOGGER.exception("Exception getting operating system pid.")
            raise

        try:
            self.policies = {}
            policy_types = list(TCellAgent.policy_classes)
            for policy_type in policy_types:
                self.policies[policy_type] = TCellAgent.policy_classes[policy_type]()
        except Exception as ex:
            LOGGER.exception("Exception loading policy types.")
            raise

        LOGGER.info("Initializing API")
        try:
            self.background_tcell_api = TCellAPI()
        except Exception as ex:
            LOGGER.exception("Exception initializing API.")
            raise

        self.shutdown = False

        LOGGER.info("Initializing IO Pipe")
        def event_callback(value):
            try:
                TCellAgent.send(value)
            except Exception as ex:
                LOGGER.debug("Exceptions handling event callback")
                LOGGER.debug(ex)
        self.fork_pipe_manager = ForkPipeManager(event_callback)
        self.fork_pipe_queue = Queue(100)
        self.fork_pipe_thread = None
        self.fork_pipe_broken = False # If the forked event send fails, start up an event manager

        def metrics_callback(value):
            try:
                _metric = value.get("_metric",None)
                LOGGER.debug("Handling callback for {metric}".format(metric=_metric))
                if (_metric == "request_metric"):
                    TCellAgent.request_metric(
                        value.get("route_id"), 
                        value.get("request_time"),
                        value.get("ip"),
                        value.get("user_agent"),
                        value.get("session_id"),
                        value.get("user_id")
                    )
                elif (_metric == "discover_database_fields"):
                    TCellAgent.discover_database_fields(
                        value.get("database"),
                        value.get("schema"),
                        value.get("table"),
                        value.get("fields"),
                        value.get("route_id")
                    )
                elif (_metric == "discover_route"):
                    TCellAgent.discover_route(
                        value.get("route_url"),
                        value.get("route_method"),
                        value.get("route_target"),
                        value.get("route_id")
                    )
                else:
                    LOGGER.debug("Unknown metric type through pipe")
                    LOGGER.debug(_metric)
            except Exception as ex:
                LOGGER.debug("Exceptions handling metrics callback")
                LOGGER.debug(ex)
        self.metrics_pipe_manger = ForkPipeManager(metrics_callback)
        self.metrics_pipe_queue = Queue(100)
        self.metrics_pipe_thread = None

        self.polling_thread = None
        self.polling_thread_pid = os.getpid()

        self.event_handling_thread = None
        LOGGER.info("Checking cache")

        self.policy_cache = TCellPolicyCache(CONFIGURATION.app_id, CONFIGURATION.cache_folder)

        LOGGER.info("Checking preload policy file")
        try:
            if CONFIGURATION.preload_policy_filename:
                self.load_policy_from_file(CONFIGURATION.preload_policy_filename)
            self.process_policies(self.policy_cache.current_cache(), cache=False)
        except Exception as ex:
            LOGGER.exception("Exception reading cache policy filename")
            raise

        LOGGER.info("Initializing route table.")
        self.route_table = RouteTable()

        self.event_queue = Queue(CONFIGURATION.event_queue_intake_size)
        self.events_to_dispatch = []
        self.agent_metrics = AgentMetrics()

    def in_fork(self):
        return (self.parent_pid != os.getpid())

    def instrument(self):
        """Call the instrumentation functions, only works if they exist"""
        from tcell_agent.instrumentation import gunicorn_tcell
        from tcell_agent.instrumentation import django
        from tcell_agent.instrumentation import flask

    def ensure_polling_thread_running(self):
        if self.is_polling_thread_running():
            return
        pollingLock.acquire()
        try:
            if self.is_polling_thread_running():
                return
            self.start_polling_thread()
        finally:
            pollingLock.release()
    def is_polling_thread_running(self):
        return self.polling_thread and self.polling_thread.isAlive() and self.polling_thread_pid == os.getpid()
    def start_polling_thread(self):
        """Start the background threads for polling/events"""
        self.polling_thread = threading.Thread(target=self.run_polling, args=())
        self.polling_thread.daemon = True # Daemonize thread
        self.polling_thread.start()                                  # Start the execution
        self.polling_thread_pid = os.getpid()
        LOGGER.info("Starting the policy-polling thread")

    def ensure_metrics_pipe_thread_running(self):
        if self.is_metrics_pipe_thread_running():
            return
        forkThreadsLock.acquire()
        try:
            if self.is_metrics_pipe_thread_running():
                return
            self.start_metrics_pipe_thread()
        finally:
            forkThreadsLock.release()
    def is_metrics_pipe_thread_running(self):
        return self.metrics_pipe_thread and self.metrics_pipe_thread.isAlive()
    def start_metrics_pipe_thread(self):
        def run_metrics_fork_thread():
            while True:
                value = self.metrics_pipe_queue.get(True)
                LOGGER.debug("Fork metrics thread received metric")
                if (self.metrics_pipe_manger.send_to_parent(value) == False):
                    self.fork_pipe_broken = True
        LOGGER.debug("Starting metrics pipe thread")
        self.metrics_pipe_thread = threading.Thread(target=run_metrics_fork_thread, args=())
        self.metrics_pipe_thread.daemon = True # Daemonize thread
        self.metrics_pipe_thread.start()        

    def ensure_fork_pipe_thread_running(self):
        if self.is_fork_pipe_thread_running():
            return
        forkThreadsLock.acquire()
        try:
            if self.is_fork_pipe_thread_running():
                return
            self.start_fork_pipe_thread()
        finally:
            forkThreadsLock.release()
    def is_fork_pipe_thread_running(self):
        return self.metrics_pipe_thread and self.metrics_pipe_thread.isAlive()
    def start_fork_pipe_thread(self):
        def run_event_fork_thread():
            while True:
                event = self.fork_pipe_queue.get(True)
                LOGGER.debug("Fork event thread received event")
                if (self.fork_pipe_manager.send_to_parent(event) == False):
                    self.fork_pipe_broken = True
        LOGGER.debug("Starting event pipe thread")
        self.fork_pipe_thread  = threading.Thread(target=run_event_fork_thread, args=())
        self.fork_pipe_thread.daemon = True # Daemonize thread
        self.fork_pipe_thread.start()  

    def ensure_event_handler_thread_running(self):
        if self.in_fork() and self.fork_pipe_broken == False:
            self.ensure_fork_pipe_thread_running()
            self.ensure_metrics_pipe_thread_running()
            return
        if self.is_event_handler_thread_running():
            return
        eventHandlerThreadsLock.acquire()
        try:
            if self.is_event_handler_thread_running():
                return
            self.start_event_handler()
        finally:
            eventHandlerThreadsLock.release()
    def is_event_handler_thread_running(self):
        return self.event_handling_thread and self.event_handling_thread.isAlive()
    def start_event_handler(self):
        self.event_handling_thread = threading.Thread(target=self.run_event_handler, args=())
        self.event_handling_thread.daemon = True # Daemonize thread
        self.event_handling_thread.start()
        LOGGER.info("Starting the event-manager thread")

    def load_policy_from_file(self, policy_filename):
        """Loads either preload policy file or cache file"""
        if os.path.isfile(policy_filename) == False:            
            LOGGER.error("Could not find the policy file: {policy_filename}".format(policy_filename=policy_filename))
            return
        with open(policy_filename) as data_file:
            policies_json = json.load(data_file)
            if policies_json.get("result"):
                policies_json = policies_json["result"]
            self.process_policies(policies_json, cache=False)

    def process_policies(self, policies_json, cache=True):
        """Given a piece of JSON (from API or file) process it"""
        policy_types = list(TCellAgent.policy_classes)
        for policy_type in policy_types:
            policy_json = policies_json.get(policy_type)
            if policy_json is not None:
                if (cache): self.policy_cache.update_cache(policy_type, policy_json)
                try:
                    self.policies[policy_type] = TCellAgent.policy_classes[policy_type](policy_json)
                except Exception as general_exception:
                    LOGGER.error("Exception parsing {policy_type} policy".format(
                        policy_type=policy_type))
                    LOGGER.exception(general_exception)

    def call_update(self, last_timestamp, delays=True):
        """Make one update request from the tCell service"""
        if CONFIGURATION.fetch_policies_from_tcell == False:
            self.shutdown = True
            return
        if CONFIGURATION.version != 0 and CONFIGURATION.version != 1:
            raise TCellAPIException("Unsupported configuration version")
        try:
            result = self.background_tcell_api.v1update(last_timestamp)
            if self.shutdown:
                return last_timestamp
            if result is not None:
                last_id = result.get("last_id", result.get("last_timestamp"))
                if last_id:
                    last_timestamp = last_id
                else:
                    return last_timestamp
                self.process_policies(result)
            else:
                LOGGER.debug("no new timestamp")
                return last_timestamp
        except TCellAPIException as tcell_api_exception:
            LOGGER.error("result exception, sleeping 30 seconds")
            LOGGER.debug(tcell_api_exception)
            if delays:
                time.sleep(30)
        except Exception as general_exception:
            LOGGER.error("result general exception, sleeping 120 seconds")
            LOGGER.debug(general_exception)
            if delays:
                time.sleep(120)
        return last_timestamp

    def run_polling(self):
        """ Method that runs forever """
        last_timestamp = None
        while self.shutdown == False:
            previous_timestamp = last_timestamp
            last_timestamp = self.call_update(last_timestamp)
            if previous_timestamp is None:
                # Just prevent two rapid requests at once.
                sleeptime = random.randint(5, 20)
                time.sleep(sleeptime)

    def addEvent(self, event):
        """Add an event to the queue to be sent to tcell"""
        if event is None:
            return
        LOGGER.debug("Adding event {event} to queue".format(event=event.get_debug_data()))
        self.ensure_event_handler_thread_running()
        try:
            if event.queue_wait == True and (self.event_queue.qsize() < CONFIGURATION.event_queue_intake_size*10):
                self.event_queue.put(event, True, 10)
            else:
                self.event_queue.put_nowait(event)
        except Full:
            pass

    def add_event_to_fork(self, event):
        LOGGER.debug("Adding event {event} to fork".format(event=event.get_debug_data()))
        self.ensure_fork_pipe_thread_running()
        try:
            self.fork_pipe_queue.put_nowait(event)
        except Full:
            pass

    def add_metric_to_fork(self, value):
        LOGGER.debug("Adding metric {value} to fork".format(value=value))
        self.ensure_metrics_pipe_thread_running()
        try:
            self.metrics_pipe_queue.put_nowait(value)
        except Full:
            pass

    def _request_metric_counter(self, route_id, request_time):
        if route_id is None or route_id == "":
            route_id = "?"
        self.agent_metrics.add_route_count(route_id, request_time)
        if (self.agent_metrics.should_send_flush_event_for_route_count_table()):
            TCellAgent.send(FlushDummyEvent())

    def run_event_handler(self):
        """ Method that runs forever """
        self.events_to_dispatch = [ServerAgentDetailsEvent(
            user=tcell_agent.system_info.current_username(),
            group=tcell_agent.system_info.current_group_id(),
            language='Python',
            language_version=system_info.python_version_string()
            )]
        self.flush_events()
        sape = ServerAgentPackagesEvent()
        def add_package_to_event(package_obj):
            sape.add_package(package_obj.key, package_obj.version)
        system_info.get_packages(add_package_to_event)
        self.events_to_dispatch = [sape]
        self.flush_events()

        last_batch_size = 0
        # Prevent one event every 59 seconds from stopping events from being sent
        last_batch_time_diff = time.time()
        while self.shutdown == False:
            try:
                time_limit = CONFIGURATION.event_time_limit_seconds
                time_since_send = time.time() - last_batch_time_diff
                time_diff = max(0, (time_limit - time_since_send))
                item = self.event_queue.get(True, time_diff)
                self.consume_event(item)
                self.event_queue.task_done()
                batch_limit = CONFIGURATION.event_batch_size_limit
                if item.flush_right_away == True:
                    last_batch_time_diff = time.time()
                    self.flush_events()
                elif len(self.events_to_dispatch) >= batch_limit:
                    last_batch_time_diff = time.time()
                    self.flush_events()
            except Empty:
                last_batch_time_diff = time.time()
                self.flush_events()
                continue

    def filter_all_ensure_events(self):
        keep_events = [event for event in self.events_to_dispatch if event.ensure_delivery == True]
        keep_events = keep_events[:200]
        self.events_to_dispatch = keep_events

    def flush_events(self):
        """Send queued events as a batch to the tCell service"""
        try:
            metrics_table = self.agent_metrics.get_and_reset_route_count_table()
            metrics_session = self.agent_metrics.get_and_reset_sesssion_metric()
            if (metrics_table or metrics_session):
                metrics_event = MetricsEvent()
                if metrics_table:
                    metrics_event.set_rct(metrics_table)
                if metrics_session:
                    metrics_event.set_session(metrics_session)
                events = self.events_to_dispatch + [metrics_event]
            else:
                events = self.events_to_dispatch
            response = self.background_tcell_api.v1send_events(events)
            if response is not None and response.status_code == 200:
                self.events_to_dispatch = []
            else:
                self.filter_all_ensure_events()
        except Exception as connection_exception:
            LOGGER.error("Error sending events, will try again laster")
            LOGGER.debug(connection_exception)
            self.filter_all_ensure_events()

    def consume_event(self, event):
        """Event is added directly, not by queue"""
        event.post_process()
        if (event.send_event):
            self.events_to_dispatch.append(event)

    @classmethod
    def send(cls, event):
        try:
            cls.get_agent().ensure_event_handler_thread_running()
            if cls.get_agent().in_fork() and cls.get_agent().fork_pipe_broken == False:
                LOGGER.debug("Sending event {event} by fork".format(event=event.get_debug_data()))
                cls.get_agent().add_event_to_fork(event)
            else:
                LOGGER.debug("Sending event {event} by queue".format(event=event.get_debug_data()))
                cls.get_agent().addEvent(event)
        except Exception as e:
            LOGGER.error("could not send event, no agent initialized")
            LOGGER.debug(e)

    @classmethod
    def get_policy(cls, policy_type):
        try:
            return cls.get_agent().policies.get(policy_type)
        except Exception as e:
            LOGGER.error("could not get policy, no agent initialized")
            LOGGER.debug(e)

    @classmethod
    def get_agent(cls):
        if cls.tCell_agent is not None:
            return cls.tCell_agent
        else:
            lock.acquire()
            try:
                if CONFIGURATION.app_id is None or CONFIGURATION.api_key is None:
                    LOGGER.error("Application ID and API Key must be set by configuration file or environmental variables.")
                    return None
                if cls.tCell_agent is not None:
                    return cls.tCell_agent
                cls.tCell_agent = TCellAgent()
            except Exception as ex:
                LOGGER.exception("Exception creating agent.")
                traceback.print_exc()
                raise
            finally:
                lock.release()
            return cls.tCell_agent

    @classmethod
    def get_route_table(cls):
        try:
            return cls.get_agent().route_table
        except Exception as e:
            LOGGER.error("could not get route_table, no agent initialized")
            LOGGER.debug(e)

    @classmethod
    def request_metric(cls, route_id, request_time, ip, user_agent, session_id=None, user_id=None):
        try:
            if cls.get_agent().in_fork() and cls.get_agent().fork_pipe_broken == False:
                LOGGER.debug("Request metric sending (in fork) {route_id} {request_time}".format(route_id=route_id,request_time=request_time))
                cls.get_agent().add_metric_to_fork({
                    "_metric":"request_metric", 
                    "route_id":route_id, 
                    "ip":ip,
                    "user_agent":user_agent,
                    "request_time":request_time,
                    "session_id":session_id,
                    "user_id":user_id
                })
            else:
                try:
                    login_policy = cls.get_policy(PolicyTypes.LOGIN)
                    if login_policy and login_policy.session_hijacking_metrics == True and user_id:
                        cls.get_agent().agent_metrics.add_session_track_metric(session_id, user_id, user_agent, ip)
                        if cls.get_agent().agent_metrics.should_send_flush_event():
                            cls.send(FlushDummyEvent())
                except Exception as exceptionAddingLoginMetric:
                    LOGGER.error("Exception adding login metric")
                    LOGGER.error(exceptionAddingLoginMetric)
                LOGGER.debug("Request metric sending {route_id} {request_time}".format(route_id=route_id,request_time=request_time))
                return cls.get_agent()._request_metric_counter(route_id, request_time)
        except Exception as e:
            LOGGER.error("could not add metric")
            LOGGER.debug(e)

    @classmethod
    def discover_route(cls, route_url, route_method, route_target, route_id):
        try:
            if cls.get_agent().in_fork() and cls.get_agent().fork_pipe_broken == False:
                cls.get_agent().add_metric_to_fork({"_metric":"discover_route", 
                    "route_url":route_url, 
                    "route_method":route_method,
                    "route_target":route_target,
                    "route_id":route_id
                })
                return
            if cls.get_agent() is not None:
                if cls.get_route_table().routes[route_id].discovered == False:
                    TCellAgent.send(AppRouteSensorEvent(
                        route_url, 
                        route_method, 
                        route_target,
                        route_id
                    ))
                    cls.get_route_table().routes[route_id].discovered = True
        except Exception as e:
            LOGGER.error("could not add route discovery")
            LOGGER.debug(e)

    @classmethod
    def discover_database_fields(cls, database, schema, table, fields, route_id=None):
        try:
            if cls.get_agent().in_fork() and cls.get_agent().fork_pipe_broken == False:
                cls.get_agent().add_metric_to_fork({"_metric":"discover_database_fields", 
                    "database":database, 
                    "schema":schema,
                    "table":table,
                    "fields":fields,
                    "route_id":route_id
                })
                return
            if cls.tCell_agent is not None:
                dlp_policy = cls.get_policy(PolicyTypes.DATALOSS)
                if dlp_policy is None or dlp_policy.database_discovery_enabled != True:
                    return
                if route_id is None:
                    route_id = ""

                if (cls.get_agent().route_table.routes[route_id].have_fields_been_discovered(database,schema,table,fields)):
                    cls.get_agent().route_table.routes[route_id].set_fields_discovered(database,schema,table,fields)
                    event = DiscoveryEvent(route_id).for_database_fields(database,schema,table,fields)
                    cls.send(event)
                    cls.get_agent().route_table.update_cache_flag = True
            else:
                return
        except Exception as e:
            LOGGER.error("could not add database discovery fields")
            LOGGER.debug(e)



# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function
from . import SensorEvent
from tcell_agent.sanitize import SanitizeUtils
from future.backports.urllib.parse import urlsplit

class AppConfigSettings(SensorEvent):
  
    def __init__(self, package, section, name, value, prefix=""):
        super(AppConfigSettings, self).__init__("app_config_setting", ensure_delivery=True, queue_wait=True)
        self.setting(package, section, prefix, name, value)
    
    def setting(self, package, section, prefix, name, value):
        self["package"] = package
        self["section"] = section
        self["prefix"] = prefix
        self["name"] = name
        self["value"] = str(value)
        



# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function
from . import SensorEvent
from tcell_agent.sanitize import SanitizeUtils
from future.backports.urllib.parse import urlsplit

class DlpEvent(SensorEvent):
    FOUND_IN_BODY = "body"
    FOUND_IN_LOG = "log"
    FOUND_IN_CONSOLE = "console"

    FRAMEWORK_VARIABLE_SESSION_ID="session_id"

    REQUEST_CONTEXT_FORM = "form"
    REQUEST_CONTEXT_COOKIE = "cookie"
    REQUEST_CONTEXT_HEADER = "header"

    def __init__(self,
        route_id,
        raw_uri,
        found_in,
        action_id=None,
        user_id=None,
        hmac_session_id=None
        ):
        super(DlpEvent, self).__init__("dlp")
        if route_id:
            self["rid"] = route_id
        self.raw_uri = raw_uri
        self["found_in"] = found_in
        if action_id:
            self["rule"] = action_id
        if user_id:
            self["uid"] = user_id
        if hmac_session_id:
            self["sid"] = hmac_session_id


    def for_database(self, database, schema, table, field):
        self["type"] = "db"
        self["db"] = database
        self["schema"] = schema
        self["table"] = table
        self["field"] = field
        return self

    def for_framework(self, framwork_variable):
        self["type"] = "framework"
        self["variable"] = framwork_variable
        return self

    def for_request(self, context, variable):
        self["type"] = "request"
        self["context"] = context
        self["variable"] = variable
        return self

    def post_process(self):
        if (self.raw_uri):
            self["uri"] = SanitizeUtils.strip_uri(self.raw_uri)



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

""" Agent Module handles communication and instrumentation, this
is the main class.
"""

from __future__ import unicode_literals
from __future__ import print_function

import threading
from queue import Queue, Empty, Full
import os
import json

from tcell_agent.instrumentation import safe_wrap_function

from tcell_agent.agent import TCellAgent, PolicyTypes
import logging
import tcell_agent.tcell_logger
LOGGER = logging.getLogger('tcell_agent').getChild(__name__)

appsensorLock = threading.Lock()

class AppSensorManager(object):
    def __init__(self):
        self._appsensor_thread = None
        self._appsensor_thread_pid = os.getpid()
        self._appsensor_queue = Queue(100)
        self.use_threads = True

    def ensure_appsensor_thread_running(self):
        if self.is_appsensor_thread_running():
            return
        appsensorLock.acquire()
        try:
            if self.is_appsensor_thread_running():
                return
            self.start_appsensor_thread()
        finally:
            appsensorLock.release()

    def is_appsensor_thread_running(self):
        return self._appsensor_thread and self._appsensor_thread.isAlive() and self._appsensor_thread_pid == os.getpid()

    def start_appsensor_thread(self):
        """Start the background threads for polling/events"""
        def run_appsensor_thread():
            while True:
                metadata = self._appsensor_queue.get(True)
                try:
                    if metadata:
                        if metadata.do_request:
                            self.run_appsensor_for_request(metadata)
                        if metadata.do_response:
                            self.run_appsensor_for_response(metadata)
                        if metadata.do_path_parameters:
                            self.run_appsensor_for_path_parameters(metadata)
                except Exception as e:
                    LOGGER.debug("Exception running appsensor")
                    LOGGER.debug(e)
        self._appsensor_thread = threading.Thread(target=run_appsensor_thread, args=())
        self._appsensor_thread.daemon = True # Daemonize thread
        self._appsensor_thread.start()
        self._appsensor_thread_pid = os.getpid()

    def run_appsensor_for_request(self, appsensor_meta):
        if appsensor_meta is None:
            return
        appsensor_policy = TCellAgent.get_policy(PolicyTypes.APPSENSOR)
        if appsensor_policy is None:
            return
        json_body = None
        if (appsensor_meta.json_body_str is not None):
            try:
                if isinstance(appsensor_meta.json_body_str, bytes):
                    json_body = json.loads(appsensor_meta.json_body_str.decode("utf-8"))
                else:
                    json_body = json.loads(appsensor_meta.json_body_str)
            except Exception as e:
                LOGGER.debug("Error decoding json")
                LOGGER.debug(e)

        appsensor_policy.run_for_request(appsensor_meta, json_body)

    def run_appsensor_for_response(self, appsensor_meta):
        if appsensor_meta is None:
            return

        appsensor_policy = TCellAgent.get_policy(PolicyTypes.APPSENSOR)
        appsensor_policy.run_for_response(appsensor_meta)

    def run_appsensor_for_path_parameters(self, appsensor_meta):
        appsensor_policy = TCellAgent.get_policy(PolicyTypes.APPSENSOR)
        if appsensor_meta is None or appsensor_policy is None:
            return

        appsensor_policy.run_for_path_parameters(appsensor_meta)

    def send_appsensor_data(self, appsensor_meta_data):
        """Add an event to the queue to be sent to tcell"""
        if appsensor_meta_data is None:
            return
        if self.use_threads == False:
            if appsensor_meta_data.do_request:
                self.run_appsensor_for_request(appsensor_meta_data)
            if appsensor_meta_data.do_response:
                self.run_appsensor_for_response(appsensor_meta_data)
            if appsensor_meta_data.do_path_parameters:
                self.run_appsensor_for_path_parameters(appsensor_meta_data)
        else:
            self.ensure_appsensor_thread_running()
            try:
                self._appsensor_queue.put_nowait(appsensor_meta_data)
            except Full:
                LOGGER.debug("Appsensor queue full")

app_sensor_manager = AppSensorManager()




from __future__ import unicode_literals
from __future__ import print_function

import unittest
from ....policies.content_security_policy import ContentSecurityPolicy
from ....policies.content_security_policy import java_hash
import re

import json

from future.backports.urllib.parse import urlparse
from future.backports.urllib.parse import parse_qs

policy_one = """
{
    "policy_id":"abc-abc-abc",
    "headers":[],
    "data": {
      "options":{
        "js_agent_api_key":"000-000-1"
      }
    }
}
"""

class ContentSecurityPolicyTest(unittest.TestCase):
    def old_header_test(self):
        policy = ContentSecurityPolicy()
        policy.loadFromHeaderString("Content-Security-Policy-Report-Only: test123")
        self.assertEqual(policy.headers(), [['Content-Security-Policy-Report-Only', ' test123']])

    def new_header_test(self):
        policy_json = {"policy_id":"xyzd", "headers":[{"name":"content-Security-POLICY", "value":"test321"}]}
        policy = ContentSecurityPolicy()
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.headers(), [['Content-Security-Policy', 'test321']])

    def header_with_report_uri_test(self):
        policy_json = {"policy_id":"xyzd", 
                       "headers":[{"name":"content-Security-policy", "value":"normalvalue","report-uri":"https://www.example.com/xys"}]
                      }
        policy = ContentSecurityPolicy()
        policy.loadFromJson(policy_json)
        policy_headers = policy.headers("1","2","3")
        policy_header = policy_headers[0]
        print(policy_header)
        self.assertEqual(policy_header[0],"Content-Security-Policy")
        policy_header_value_parts = urlparse(policy_header[1].split(" ")[2])
        self.assertEqual(policy_header_value_parts.path, "/xys")
        print(policy_header_value_parts.query)
        query_params = parse_qs(policy_header_value_parts.query)
        print(query_params)
        self.assertEqual(query_params["tid"], ["1"])
        self.assertEqual(query_params["rid"], ["2"])
        self.assertEqual(query_params["sid"], ["3"])
        fromstr = policy_header[1].split(" ")[2]
        check_against = policy.policy_id + re.match("(.*)&c=[a-zA-Z0-9\-]+$", fromstr).group(1)
        self.assertEqual(query_params["c"], [str(java_hash(check_against))])

    def check_java_hash_test(self):
        base_str = "What the heck?"
        self.assertEqual(str(java_hash(base_str)), "277800975")
        self.assertEqual(str(java_hash("https://example.com/abcdde?tid=1&sid=3&rid=2")), "947858466")
        self.assertEqual(str(java_hash("Hello World")), "-862545276")

    def handle_js_agent_add(self):
        policy_json = json.loads(policy_one)
        policy = ContentSecurityPolicy()
        self.assertEqual(policy.js_agent_api_key, "000-000-1")




# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

from tcell_agent.agent import TCellAgent, PolicyTypes
from tcell_agent.policies import DataLossPolicy
from tcell_agent.sensor_events import DlpEvent
import cgi
import re
from collections import defaultdict


class ContextFilter(object):
    DATABASE = "db"
    REQUEST = "request"
    def __init__(self):
        self.type = None
        self.rule = None
        self.context = None
        self.parameter = None
        self.database = None
        self.schema = None
        self.table = None
        self.field = None
    def for_request(self, context, parameter, rule):
        self.type = ContextFilter.REQUEST
        self.context = context
        self.parameter = parameter
        self.rule =rule
        return self
    def for_database(self, database, schema, table, field, rule):
        self.type = ContextFilter.DATABASE
        self.database = database
        self.schema = schema
        self.table = table
        self.field = field
        self.rule = rule
        return self
    def __hash__(self):
        return hash("{0},{1},{2},{3},{4},{5},{6},{7}".format(
            self.type,
            self.rule,
            self.parameter,
            self.database,
            self.schema,
            self.table,
            self.field,
            self.context
            ))

class TCellInstrumentationContext(object):
    def __init__(self):
        self.filter_from_body = set() 
        self.filter_from_log = set()
        self.raw_session_id = None
        self.session_id = None # HMAC'd session id
        self.user_id = None
        self.user_agent = None
        self.remote_addr = None
        self.transaction_id = None
        self.route_id = None
        self.fullpath = None
        self.filters = defaultdict(set)
        self.start_time = 0

    def validate_term(self, term):
        if (term is not None and term != '' and term != DataLossPolicy.B_REDACT_MESSAGE and term != DataLossPolicy.REDACT_MESSAGE):
            if (len(str(term)) > 4):
                return True
        return False

    def add_response_db_filter(self, term, action_obj, database, schema, table, field):
        try:
            if self.validate_term(term) == False:
                return
            self.filters[re.escape(str(term))].add(ContextFilter().for_database(database,schema,table,field,action_obj))
        except Exception as e:
            pass

    def add_response_parameter_filter(self, term, action_obj, parameter_name):
        try:
            if self.validate_term(term) == False:
                return
            self.filters[re.escape(str(term))].add( ContextFilter().for_request("form",parameter_name,action_obj) )
        except Exception as e:
            pass

    def add_response_cookie_filter(self, term, action_obj, parameter_name):
        try:
            if self.validate_term(term) == False:
                return
            self.filters[re.escape(str(term))].add( ContextFilter().for_request("cookie",parameter_name,action_obj) )
        except Exception as e:
            pass

    def add_header_filter(self, term, action_obj, header_name):
        try:
            if self.validate_term(term) == False:
                return
            self.filters[re.escape(str(term))].add( ContextFilter().for_request("header",header_name,action_obj) )
        except Exception as e:
            print(e)
            pass

    def add_response_body_filter(self, string):
        self.filter_from_body.add(string)
        self.filter_from_body.add(cgi.escape(string))

    def add_log_filter(self, string):
        self.filter_from_log.add(string)

    def filter_body(self, body):
        dataloss_policy = TCellAgent.get_policy(PolicyTypes.DATALOSS)
        if dataloss_policy: 
            def filterx(body, replace, send, term):
                send_event_flags = {"send":False}
                def pythong_2_or_3_repl():
                    if isinstance(body, str) == False:
                        return DataLossPolicy.B_REDACT_MESSAGE
                    return DataLossPolicy.REDACT_MESSAGE       
                def repl(found_term):
                    if replace:
                        send_event_flags["send"] = True
                        return pythong_2_or_3_repl()
                    if send:
                        send_event_flags["send"] = True
                    return found_term.group(0)
                if isinstance(body, str) == False and isinstance(term, str):
                    encoded_term = term.encode("utf-8")
                else:
                    encoded_term = term
                try:
                    body = re.sub(encoded_term, repl, body)
                except:
                    pass
                return (send_event_flags["send"], body)
            if (len(self.filters) > 0):
                terms = [term for term in self.filters]
                terms.sort(key=len, reverse=True)
                for term in terms:
                    context_filters = self.filters[term]
                    replace = len([ai for ai in context_filters if ai.rule.body_redact==True]) > 0
                    send = len([ai for ai in context_filters if ai.rule.body_event==True]) > 0
                    send_event, body = filterx(body, replace, send, term)
                    if send_event:
                        for context_filter in context_filters:
                            if (context_filter.type == ContextFilter.DATABASE):
                                dlpe = DlpEvent(self.route_id, self.fullpath, DlpEvent.FOUND_IN_BODY, context_filter.rule.id).for_database(
                                    context_filter.database, 
                                    context_filter.schema, 
                                    context_filter.table, 
                                    context_filter.field)
                                TCellAgent.send(dlpe)
                            elif (context_filter.type == ContextFilter.REQUEST):
                                dlpe = DlpEvent(self.route_id, self.fullpath, DlpEvent.FOUND_IN_BODY, context_filter.rule.id).for_request(
                                    context_filter.context,
                                    context_filter.parameter)
                                TCellAgent.send(dlpe)                             
            if self.raw_session_id:
               session_id_actions = dataloss_policy.get_actions_for_session_id()
               if session_id_actions:
                   send_event, body = filterx(body, session_id_actions.body_redact, session_id_actions.body_event, self.raw_session_id)
                   if send_event:
                       dlpe = DlpEvent(self.route_id, self.fullpath, DlpEvent.FOUND_IN_BODY).for_framework(DlpEvent.FRAMEWORK_VARIABLE_SESSION_ID)
                       TCellAgent.send(dlpe)

        return body

    def filter_log_message(self, msg):
        dataloss_policy = TCellAgent.get_policy(PolicyTypes.DATALOSS)
        if dataloss_policy:
            def filterx(body, replace, send, term):
                send_event_flags = {"send":False}
                def pythong_2_or_3_repl():
                    if isinstance(body, str) == False:
                        return DataLossPolicy.B_REDACT_MESSAGE
                    return DataLossPolicy.REDACT_MESSAGE 
                def repl(found_term):
                    if replace:
                        send_event_flags["send"] = True
                        return pythong_2_or_3_repl()
                    if send:
                        send_event_flags["send"] = True
                    return found_term.group(0)
                if isinstance(body, str) == False and isinstance(term, str):
                    encoded_term = term.encode("utf-8")
                else:
                    encoded_term = term
                try:
                    body = re.sub(encoded_term, repl, body)
                except:
                    pass
                return (send_event_flags["send"], body)

            terms = [term for term in self.filters]
            terms.sort(key=len, reverse=True)
            for term in terms:
                context_filters = self.filters[term]
                replace = len([ai for ai in context_filters if ai.rule.log_redact==True]) > 0
                send = len([ai for ai in context_filters if ai.rule.log_event==True]) > 0
                send_event, msg = filterx(msg, replace, send, term)
                if send_event:
                    for context_filter in context_filters:
                        if (context_filter.type == ContextFilter.DATABASE):
                            dlpe = DlpEvent(self.route_id, self.fullpath, DlpEvent.FOUND_IN_LOG, context_filter.rule.id).for_database(
                                context_filter.database, 
                                context_filter.schema, 
                                context_filter.table, 
                                context_filter.field)
                            TCellAgent.send(dlpe)
                        elif (context_filter.type == ContextFilter.REQUEST):
                            dlpe = DlpEvent(self.route_id, self.fullpath, DlpEvent.FOUND_IN_LOG, context_filter.rule.id).for_request(
                                context_filter.context,
                                context_filter.parameter)
                            TCellAgent.send(dlpe)      
            session_id_actions = dataloss_policy.get_actions_for_session_id()
            if session_id_actions:
                send_event, msg = filterx(msg, session_id_actions.log_redact, session_id_actions.log_event, self.raw_session_id)
                if send_event:
                    dlpe = DlpEvent(self.route_id, self.fullpath, DlpEvent.FOUND_IN_LOG).for_framework(DlpEvent.FRAMEWORK_VARIABLE_SESSION_ID)
                    TCellAgent.send(dlpe)
        return msg



from __future__ import unicode_literals
from __future__ import print_function

import unittest

from mock import Mock, patch
from tcell_agent.appsensor.meta import AppSensorMeta
from tcell_agent.policies.appsensor import SqliSensor

class SqliSensorTest(unittest.TestCase):

    def create_default_sensor_test(self):
        sensor = SqliSensor()
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.exclude_headers, False)
        self.assertEqual(sensor.exclude_forms, False)
        self.assertEqual(sensor.exclude_cookies, False)
        self.assertEqual(sensor.dp, "sqli")
        self.assertEqual(sensor.libinjection, False)
        self.assertEqual(sensor.active_pattern_ids, {})
        self.assertEqual(sensor.exclusions, {})

    def create_enabled_sensor_test(self):
        sensor = SqliSensor({"enabled": True})
        self.assertEqual(sensor.enabled, True)
        self.assertEqual(sensor.exclude_headers, False)
        self.assertEqual(sensor.exclude_forms, False)
        self.assertEqual(sensor.exclude_cookies, False)
        self.assertEqual(sensor.dp, "sqli")
        self.assertEqual(sensor.libinjection, False)
        self.assertEqual(sensor.active_pattern_ids, {})
        self.assertEqual(sensor.exclusions, {})

    def create_excluded_headers_sensor_test(self):
        sensor = SqliSensor({"exclude_headers": True})
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.exclude_headers, True)
        self.assertEqual(sensor.exclude_forms, False)
        self.assertEqual(sensor.exclude_cookies, False)
        self.assertEqual(sensor.dp, "sqli")
        self.assertEqual(sensor.libinjection, False)
        self.assertEqual(sensor.active_pattern_ids, {})
        self.assertEqual(sensor.exclusions, {})

    def create_excluded_forms_sensor_test(self):
        sensor = SqliSensor({"exclude_forms": True})
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.exclude_headers, False)
        self.assertEqual(sensor.exclude_forms, True)
        self.assertEqual(sensor.exclude_cookies, False)
        self.assertEqual(sensor.dp, "sqli")
        self.assertEqual(sensor.libinjection, False)
        self.assertEqual(sensor.active_pattern_ids, {})
        self.assertEqual(sensor.exclusions, {})

    def create_excluded_cookies_sensor_test(self):
        sensor = SqliSensor({"exclude_cookies": True})
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.exclude_headers, False)
        self.assertEqual(sensor.exclude_forms, False)
        self.assertEqual(sensor.exclude_cookies, True)
        self.assertEqual(sensor.dp, "sqli")
        self.assertEqual(sensor.libinjection, False)
        self.assertEqual(sensor.active_pattern_ids, {})
        self.assertEqual(sensor.exclusions, {})

    def create_libinjection_sensor_test(self):
        sensor = SqliSensor({"libinjection": True})
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.exclude_headers, False)
        self.assertEqual(sensor.exclude_forms, False)
        self.assertEqual(sensor.exclude_cookies, False)
        self.assertEqual(sensor.dp, "sqli")
        self.assertEqual(sensor.libinjection, True)
        self.assertEqual(sensor.active_pattern_ids, {})
        self.assertEqual(sensor.exclusions, {})

    def create_active_pattern_ids_sensor_test(self):
        sensor = SqliSensor({"patterns": ["1", "2", "3"]})
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.exclude_headers, False)
        self.assertEqual(sensor.exclude_forms, False)
        self.assertEqual(sensor.exclude_cookies, False)
        self.assertEqual(sensor.dp, "sqli")
        self.assertEqual(sensor.libinjection, False)
        self.assertEqual(sensor.active_pattern_ids, {"1": True, "2": True, "3": True})
        self.assertEqual(sensor.exclusions, {})

    def create_exclusions_sensor_test(self):
        sensor = SqliSensor({"exclusions": {"word": ["form", "header"]}})
        self.assertEqual(sensor.enabled, False)
        self.assertEqual(sensor.exclude_headers, False)
        self.assertEqual(sensor.exclude_forms, False)
        self.assertEqual(sensor.exclude_cookies, False)
        self.assertEqual(sensor.dp, "sqli")
        self.assertEqual(sensor.libinjection, False)
        self.assertEqual(sensor.active_pattern_ids, {})
        self.assertEqual(sensor.exclusions, {"word": ["form", "header"]})

    def with_disabled_sensor_check_test(self):
        sensor = SqliSensor({"enabled": False})

        payloads_policy = Mock()
        with patch('tcell_agent.policies.appsensor.injection_sensor.sendEvent') as patched_send_event:
            sensor.check("get", {}, None, None, payloads_policy)
            self.assertFalse(patched_send_event.called)
            self.assertFalse(payloads_policy.apply.called)

    def with_enabled_sensor_and_no_vuln_params_check_test(self):
        sensor = SqliSensor({"enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )

        payloads_policy = Mock()
        with patch('tcell_agent.policies.appsensor.injection_sensor.sendEvent') as patched_send_event:
            sensor.check("get", meta, None, None, payloads_policy)
            self.assertFalse(patched_send_event.called)
            self.assertFalse(payloads_policy.apply.called)

    def with_enabled_sensor_and_vuln_params_check_test(self):
        sensor = SqliSensor({"enabled": True, "patterns": ["1"]})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )

        payloads_policy = Mock()
        payloads_policy.apply.return_value = None
        with patch('tcell_agent.policies.appsensor.injection_sensor.sendEvent') as patched_send_event:
            is_sqli = sensor.check("get", meta, "dirty", "Erwin' OR '1'='1", payloads_policy)
            self.assertTrue(is_sqli)
            patched_send_event.assert_called_once_with(meta, "sqli",  'dirty', {"l": "query"}, None, "1")
            payloads_policy.apply.assert_called_once_with("sqli", meta, "get", "dirty", "Erwin' OR '1'='1", {"l": "query"}, "1")



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from builtins import bytes

from future.backports.urllib.parse import urlencode
from future.backports.urllib.parse import urlsplit
from future.backports.urllib.parse import urlunsplit
from future.backports.urllib.parse import parse_qs

from tcell_agent.config import CONFIGURATION
from http.cookies import BaseCookie

import hmac
import hashlib
import sys

class SanitizeUtils:
    radius = 0

    @classmethod
    def hmac(cls, value):
        hmac_key = "hmac_key"
        if (CONFIGURATION):
            if (CONFIGURATION.hmac_key is not None):
                hmac_key = CONFIGURATION.hmac_key
            elif (CONFIGURATION.app_id is not None):
                hmac_key = CONFIGURATION.app_id
        try:
            hmac_key_enc = bytes(hmac_key, "utf-8")
        except:
            hmac_key_enc = bytes(hmac_key)
        digest = hmac.new(hmac_key_enc, str(value).encode('utf-8'), hashlib.sha256).hexdigest()
        return digest

    @classmethod
    def hmac_half(cls, value):
        hmacd = cls.hmac(value)
        if hmacd and len(hmacd) > 40:
            hmacd = hmacd[0:40]
        return hmacd

    @classmethod
    def strip_query_string(cls, query_string):
        query_string_type = str(type(query_string))

        # use utf-8 decoding to change it to unicode in python2
        if query_string_type.startswith("<type 'str'"):
            query_string = query_string.decode('utf-8')

        query_params = parse_qs(query_string)

        for param_name, param_values in query_params.items():
            new_param_values = []
            for param_value in param_values:
                new_param_values.append("")
            query_params[param_name] = new_param_values
        new_query_string = urlencode(query_params, doseq=True)
        return new_query_string

    @classmethod
    def sanitize_query_string(cls, query_string):
        query_params = parse_qs(query_string)
        blacklist = ["passwd","token","password","key"]
        for param_name, param_values in query_params.items():
            new_param_values = []
            for param_value in param_values:
                if param_name in blacklist:
                    new_param_values.append("x")
                    continue
                new_param_values.append( cls.hmac(param_value) )
            query_params[param_name] = new_param_values
        new_query_string = urlencode(query_params, doseq=True)
        return new_query_string

    @classmethod
    def sanitize_uri(cls, uri):
        scheme, netloc, path, query_string, fragment = urlsplit(uri)
        query_string = cls.sanitize_query_string(query_string)
        if fragment is not None and fragment != "":
            fragment = "x"
        return urlunsplit((str(scheme), str(netloc), str(path), str(query_string), str(fragment)))

    @classmethod
    def strip_uri(cls, uri):
        scheme, netloc, path, query_string, fragment = urlsplit(uri)
        query_string = cls.strip_query_string(query_string)
        if fragment is not None and fragment != "":
            fragment = cls.hmac(fragment)

        return urlunsplit((str(scheme), str(netloc), str(path), str(query_string), str(fragment)))

    @classmethod
    def sanitize_request_info(cls, request_info):
        headers = request_info.get("headers")
        if headers:
            sanitized_headers = {}
            for header_name, header_values in headers.items():
                sanitized_headers[header_name] = []
                header_name_l = header_name.lower()
                if header_name_l == "cookie":
                    for header_value in header_values:
                        base_cookie = BaseCookie(str(header_value));
                        for cookie_name, cookie_value in base_cookie.items():
                            base_cookie[cookie_name] = cls.hmac(cookie_value.coded_value)
                        sanitized_headers[header_name].append( base_cookie.output(header="", sep=';') )
                elif header_name_l == "referer":
                    for header_value in header_values:
                        sanitized_headers[header_name].append(cls.sanitize_uri(header_value))
                else:
                    sanitized_headers[header_name] = header_values
            request_info["headers"] = sanitized_headers
        uri = request_info["uri"]
        if uri:
            request_info["uri"] = cls.sanitize_uri(uri)
        post_data = request_info.get("post_data")
        if post_data:
            request_info["post_data"] = cls.sanitize_query_string(post_data)
        return request_info

    @classmethod
    def sanitize_response_info(cls, response_info):
        headers = response_info.get("headers")
        if headers:
            sanitized_headers = {}
            for header_name, header_values in headers.items():
                sanitized_headers[header_name] = []
                header_name_l = header_name.lower()
                if header_name_l == "set-cookie":
                    for header_value in header_values:
                        base_cookie = BaseCookie(str(header_value));
                        for cookie_name, cookie_value in base_cookie.items():
                            base_cookie[cookie_name] = cls.hmac(cookie_value.coded_value)
                        sanitized_headers[header_name].append( base_cookie.output(header="", sep=';').strip() )
                else:
                    sanitized_headers[header_name] = header_values
            response_info["headers"] = sanitized_headers
        return response_info



from __future__ import unicode_literals
from __future__ import print_function

import unittest

from django.db.utils import ProgrammingError, InterfaceError

from mock import patch, Mock
from tcell_agent.appsensor.meta import AppSensorMeta
from tcell_agent.policies.appsensor import MiscSensor

class MiscSensorTest(unittest.TestCase):

    def create_default_sensor_test(self):
        sensor = MiscSensor()
        self.assertFalse(sensor.csrf_exception_enabled)
        self.assertFalse(sensor.sql_exception_enabled)

    def create_enabled_sensor_test(self):
        sensor = MiscSensor({
            "csrf_exception_enabled": True,
            "sql_exception_enabled": True
        })
        self.assertTrue(sensor.csrf_exception_enabled)
        self.assertTrue(sensor.sql_exception_enabled)

    def with_disabled_sensor_csrf_rejected_test(self):
        sensor = MiscSensor({"csrf_exception_enabled": False})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.misc_sensor.sendEvent') as patched_send_event:
            sensor.csrf_rejected(meta, "Missing CSRF")
            self.assertFalse(patched_send_event.called)

    def with_enabled_sensor_csrf_rejected_test(self):
        sensor = MiscSensor({"csrf_exception_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.misc_sensor.sendEvent') as patched_send_event:
            sensor.csrf_rejected(meta, "Missing CSRF")
            patched_send_event.assert_called_once_with(meta, "excsrf", None, None)

    def with_disabled_sensor_sql_exception_test(self):
        sensor = MiscSensor({"sql_exception_enabled": False})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.misc_sensor.sendEvent') as patched_send_event:
            with patch('traceback.print_tb', return_value="printed stack") as patched_print_tb:
                db = Mock()
                meta = Mock()
                exc_type = Mock()
                exc_value = "some error"
                tb = Mock()

                sensor.sql_exception_detected(db, meta,  exc_type, exc_value, tb)
                self.assertFalse(patched_send_event.called)
                self.assertFalse(patched_print_tb.called)

    def with_enabled_sensor_sql_exception_test(self):
        sensor = MiscSensor({"sql_exception_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.misc_sensor.sendEvent') as patched_send_event:
            with patch('traceback.print_tb', return_value="printed stack") as patched_print_tb:
                db = Mock()
                db.ProgrammingError = ProgrammingError
                exc_type = ProgrammingError
                exc_value = "some error"
                tb = Mock()

                sensor.sql_exception_detected(db, meta,  exc_type, exc_value, tb)
                patched_send_event.assert_called_once_with(meta, "exsql", "ProgrammingError", None, payload='printed stack')
                patched_print_tb.assert_called_once_with(tb)

    def with_enabled_sensor_sql_exception_interface_error_test(self):
        sensor = MiscSensor({"sql_exception_enabled": True})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.misc_sensor.sendEvent') as patched_send_event:
            with patch('traceback.print_tb', return_value="printed stack") as patched_print_tb:
                db = Mock()
                db.ProgrammingError = ProgrammingError
                exc_type = InterfaceError
                exc_value = "some error"
                tb = Mock()

                sensor.sql_exception_detected(db, meta,  exc_type, exc_value, tb)
                self.assertFalse(patched_send_event.called)
                self.assertFalse(patched_print_tb.called)

    def with_enabled_sensor_sql_exception_matching_excluded_route_test(self):
        sensor = MiscSensor({"sql_exception_enabled": True, "exclude_routes": ["23947"]})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.misc_sensor.sendEvent') as patched_send_event:
            with patch('traceback.print_tb', return_value="printed stack") as patched_print_tb:
                db = Mock()
                db.ProgrammingError = ProgrammingError
                exc_type = ProgrammingError
                exc_value = "some error"
                tb = Mock()

                sensor.sql_exception_detected(db, meta,  exc_type, exc_value, tb)
                self.assertFalse(patched_send_event.called)
                self.assertFalse(patched_print_tb.called)

    def with_enabled_sensor_sql_exception_nonmatching_excluded_route_test(self):
        sensor = MiscSensor({"sql_exception_enabled": True, "exclude_routes": ["nonmatching"]})
        meta = AppSensorMeta(
            "remote_addr",
            "request_method",
            "abosolute_uri",
            "23947",
            session_id="session_id",
            user_id="user_id"
        )
        with patch('tcell_agent.policies.appsensor.misc_sensor.sendEvent') as patched_send_event:
            with patch('traceback.print_tb', return_value="printed stack") as patched_print_tb:
                db = Mock()
                db.ProgrammingError = ProgrammingError
                exc_type = ProgrammingError
                exc_value = "some error"
                tb = Mock()

                sensor.sql_exception_detected(db, meta,  exc_type, exc_value, tb)
                patched_send_event.assert_called_once_with(meta, "exsql", "ProgrammingError", None, payload='printed stack')
                patched_print_tb.assert_called_once_with(tb)



import json

from tcell_agent.policies.appsensor.sensor import sendEvent

class DatabaseSensor(object):
    DP_CODE="dbmaxrows"

    def __init__(self, policy_json=None):
        self.enabled = False
        self.max_rows = 1001
        self.excluded_route_ids = {}

        if policy_json is not None:
            self.enabled = policy_json.get("enabled", False)
            large_result = policy_json.get("large_result", {})
            self.max_rows = large_result.get("limit", self.max_rows)

            for route_id in policy_json.get("exclude_routes", []):
                self.excluded_route_ids[route_id] = True

    def check(self, appsensor_meta, number_of_records):
        if (not self.enabled) or self.excluded_route_ids.get(appsensor_meta.route_id, False):
            return

        if number_of_records > self.max_rows:
            sendEvent(
              appsensor_meta,
              self.DP_CODE,
              None,
              { "rows": number_of_records })

    def __str__(self):
        return "<%s enabled: %s max_rows: %s dp_code: %s excluded_route_ids: %s>" % \
            (type(self).__name__, self.enabled, self.max_rows, self.DP_CODE, self.excluded_route_ids)



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

from tcell_agent.policies import TCellPolicy

class HttpTxPolicy(TCellPolicy):
    def __init__(self, policy_json=None):
        super(HttpTxPolicy, self).__init__()
        self.firehose = { "enabled": False, "lite":False }
        self.auth_framework = { "enabled": False, "lite":False }
        self.profile = { "enabled": False }
        self.fingerprint = { "enabled": False }

        if policy_json is not None:
            self.loadFromJson(policy_json)

    def loadFromJson(self, policy_json):
        if "policy_id" in policy_json:
            self.policy_id = policy_json["policy_id"]
        else:
            raise Exception("Policy Id Not Found")

        self.firehose = { "enabled": False, "lite":False }
        self.auth_framework = { "enabled": False, "lite":False }
        self.profile = { "enabled": False }
        self.fingerprint = { "enabled": False, "hmacUserAgent":False, "hmacUserId":False, "sampling":0 }

        types = policy_json.get("types")
        if types:
            firehose_settings = types.get("firehose")
            if firehose_settings:
                self.firehose["enabled"] = firehose_settings.get("enabled", False)        
                self.firehose["lite"] = firehose_settings.get("lite", False)        
            if types.get("auth_framework"):
                self.auth_framework["enabled"] = types.get("auth_framework").get("enabled", False)
                self.auth_framework["lite"] = types.get("auth_framework").get("lite", False)
            if types.get("profile"):
                self.profile["enabled"] = types.get("profile").get("enabled", False)
            if types.get("fingerprint"):
                self.fingerprint["enabled"] = types.get("fingerprint").get("enabled", False)
                self.fingerprint["hmacUserAgent"] = types.get("fingerprint").get("hmacUserAgent", False)
                self.fingerprint["hmacUserId"] = types.get("fingerprint").get("hmacUserId", False)
                self.fingerprint["sampling"] = types.get("fingerprint").get("sampling", 0)



from tcell_agent.agent import TCellAgent, PolicyTypes
from tcell_agent.instrumentation.manager import InstrumentationManager
from tcell_agent.appsensor.django import django_meta


def instrument_csrf_view_middleware():
    from django.middleware.csrf import CsrfViewMiddleware

    def _tcell_reject(_tcell_original_reject, self, request, reason):
        appsensor_policy = TCellAgent.get_policy(PolicyTypes.APPSENSOR)
        if appsensor_policy is not None:
            meta = django_meta(request)
            appsensor_policy.csrf_rejected(meta, reason)

        return _tcell_original_reject(self, request, reason)
    InstrumentationManager.instrument(CsrfViewMiddleware, "_reject", _tcell_reject)



from __future__ import unicode_literals
from __future__ import print_function
import unittest
from ....policies.secure_headers_policy import SecureHeadersPolicy

class SecureHeaderPolicyTest(unittest.TestCase):
    def reserved_header_test(self):
        policy_json = {"policy_id":"xyzd", "headers":[{"name":"content-Security-POLICY", "value":"test321"}]}
        policy = SecureHeadersPolicy()
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.headers(), [])

    def one_header_test(self):
        policy_json = {"policy_id":"xyzd", "headers":[{"name":"X-Content-Type-Options", "value":"nosniff"}]}
        policy = SecureHeadersPolicy()
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.headers(), [["X-Content-Type-Options","nosniff"]])



# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function
import getpass
import pip
import os
import sys

try:
  import pwd
except ImportError:
  import getpass
  pwd = None
import errno

def mkdir_p(path):
    """http://stackoverflow.com/a/600612/190597 (tzot)"""
    try:
        os.makedirs(path, exist_ok=True)  # Python>3.2
    except TypeError:
        try:
            os.makedirs(path)
        except OSError as exc: # Python >2.5
            if exc.errno == errno.EEXIST and os.path.isdir(path):
                pass
            else: raise

def python_version_string():
  try:
    return sys.version.split("\n")[0]
  except:
    return None

def current_username():
  if pwd: # not available on windows
    return pwd.getpwuid(os.geteuid()).pw_name
  else:
    return getpass.getuser()

def current_user_id():
  return os.getuid()

def current_process_id():
  return os.getpid()

def current_group_id():
  return str(os.getegid()) # effective group id

def get_packages(callback=None):
    installed_packages = pip.get_installed_distributions()
    for installed_package in installed_packages:
        if callback:
            callback(installed_package)
    #installed_packages_list = sorted(["%s==%s" % (i.key, i.version) for i in installed_packages])
    return installed_packages


# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

from tcell_agent.sanitize import SanitizeUtils

from . import SensorEvent

class RedirectSensorEvent(SensorEvent):
    def __init__(self,
                 remote_addr,
                 method,
                 from_domain,
                 from_full_path,
                 status_code,
                 redirect_url,
                 user_id=None,
                 hmac_session_id=None,
                 route_id=None):
        super(RedirectSensorEvent, self).__init__("redirect")
        self["method"] = method
        self["remote_addr"] = remote_addr
        self["from_domain"] = from_domain
        self.raw_full_path = from_full_path
        self["status_code"] = status_code
        self["to"] = redirect_url
        self.raw_user_id = user_id
        self.hmac_session_id = hmac_session_id
        if route_id:
            self["rid"] = route_id

    def post_process(self):
        self["from"] = SanitizeUtils.strip_uri(self.raw_full_path)

        if self.raw_user_id:
            self["uid"] = self.raw_user_id
        if self.hmac_session_id:
            self["sid"] = self.hmac_session_id



# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function
import json
import re
from tcell_agent.sanitize import SanitizeUtils
from . import SensorEvent
import logging
logger = logging.getLogger('tcell_agent').getChild(__name__)

class HttpTxSensorEvent(SensorEvent):
    def __init__(self, request, response, event_name=None):
        super(HttpTxSensorEvent, self).__init__("http_tx")
        if event_name is not None:
            self["event_name"] = event_name
        self.raw_request = request
        self.raw_response = response

    def post_process(self):
        self["request"] = {}
        self["response"] = {}
        try:
            from django.http.request import HttpRequest
            from django.http.response import HttpResponse
            if isinstance(self.raw_request, HttpRequest):
                self["request"] = HttpRequestInfo.from_django_request(self.raw_request)
            if isinstance(self.raw_response, HttpResponse):
                self["response"] = HttpResponseInfo.from_django_response(self.raw_response)
        except Exception as e:
            # Not Django-request-response
            logger.debug(e)
        self["request"] = SanitizeUtils.sanitize_request_info(self["request"])
        self["response"] = SanitizeUtils.sanitize_response_info(self["response"])

class LoginFailureSensorEvent(HttpTxSensorEvent):
    def __init__(self, request, response):
        super(LoginFailureSensorEvent, self).__init__(request, response)
        self["event_name"] = "login_failure"

class LoginSuccessfulSensorEvent(HttpTxSensorEvent):
    def __init__(self, request, response):
        super(LoginSuccessfulSensorEvent, self).__init__(request, response)
        self["event_name"] = "login_success"

class FingerprintSensorEvent(SensorEvent):
    def __init__(self, request, hmac_session_id, user_id=None):
        super(FingerprintSensorEvent, self).__init__("fingerprint")
        self.hmac_session_id = hmac_session_id
        self.raw_request = request
        self.user_id = user_id
    def post_process(self):
        request = None
        if (self.raw_request._tcell_sensors_httpx.get("uuid")):
            self["tid"] = self.raw_request._tcell_sensors_httpx.get("uuid")
        try:
            from django.http.request import HttpRequest
            from django.http.response import HttpResponse
            if isinstance(self.raw_request, HttpRequest):
                request = HttpRequestInfo.from_django_request(self.raw_request)
        except Exception as e:
            # Not Django-request-response
            logger.debug(e)
        if request:
            # we have a valid request/response pair
            ua = request.get("headers",{}).get("User_agent")
            if ua:
                ua = "".join(ua)
            ip = request.get("remote_host")
            sid = self.hmac_session_id
            self["ua"] = ua
            self["ip"] = ip
            self["sid"] = sid
            if self.user_id:
                self["uid"] = self.user_id

class HttpMessageInfo(dict):
    def __init__(self):
        self["headers"] = {}

class HttpRequestInfo(HttpMessageInfo):
    uri = None
    method = None
    remote_host = None
    post_data = None
    def __str__(self):
        return json.dumps(self.__dict__)
    @classmethod 
    def from_django_request(cls, request):
        request_info = HttpRequestInfo()
        request_info["uri"] = request.build_absolute_uri()
        request_info["remote_host"] = request.META.get("REMOTE_ADDR")
        request_info["method"] = request.method
        request_info["post_data"] = request.body
        regex = re.compile('^HTTP_')
        raw_headers = dict((regex.sub('', header), value) for (header, value) 
               in request.META.items() if header.startswith('HTTP_'))
        headers = {}
        for header_name, header_value in raw_headers.items():
            header_name = header_name.capitalize()
            if headers.get(header_name):
                headers[header_name].append(header_value)
            else:
                headers[header_name] = [header_value]
        request_info["headers"] = headers
        return request_info

class HttpResponseInfo(HttpMessageInfo):
    def __str__(self):
        return json.dumps(self.__dict__)
    @classmethod 
    def from_django_response(cls, response):
        response_info = HttpResponseInfo()
        response_info["status"] = response.status_code
        response_info["content_type"] = response.get("content-type")
        headers = {}
        for header_name, header_value in response.items():
            if headers.get(header_name):
                headers[header_name].append(header_value)
            else:
                headers[header_name] = [header_value]
        response_info["headers"] = headers
        return response_info



from __future__ import unicode_literals
from __future__ import print_function

import json
import unittest

from nose.tools import raises

from tcell_agent.policies import PatchesPolicy

class AppSensorPolicyTest(unittest.TestCase):
    def classname_test(self):
        self.assertEqual(PatchesPolicy.api_identifier, "patches")

    def none_policy_test(self):
        policy = PatchesPolicy()

        self.assertIsNone(policy.policy_id)
        self.assertIsNone(policy.version)
        self.assertFalse(policy.ip_blocking_enabled)
        self.assertEqual(policy.blocked_ips, {})

    @raises(Exception)
    def empty_policy_test(self):
        policy = PatchesPolicy({})

    def empty_version_policy_test(self):
        policy = PatchesPolicy({
            "policy_id": "policy_id"
        })

        self.assertEqual(policy.policy_id, "policy_id")
        self.assertIsNone(policy.version)
        self.assertFalse(policy.ip_blocking_enabled)
        self.assertEqual(policy.blocked_ips, {})

    def empty_data_policy_test(self):
        policy = PatchesPolicy({
            "policy_id": "policy_id",
            "version": 1,
            "data": {}
        })

        self.assertEqual(policy.policy_id, "policy_id")
        self.assertEqual(policy.version, 1)
        self.assertFalse(policy.ip_blocking_enabled)
        self.assertEqual(policy.blocked_ips, {})

    def empty_blocked_ips_policy_test(self):
        policy = PatchesPolicy({
            "policy_id": "policy_id",
            "version": 1,
            "data": {
                "blocked_ips": []
            }
        })

        self.assertEqual(policy.policy_id, "policy_id")
        self.assertEqual(policy.version, 1)
        self.assertFalse(policy.ip_blocking_enabled)
        self.assertEqual(policy.blocked_ips, {})

    def populated_blocked_ips_policy_test(self):
        policy = PatchesPolicy({
            "policy_id": "policy_id",
            "version": 1,
            "data": {
                "blocked_ips": [{
                    "ip": "1.1.1.1"
                }]
            }
        })

        self.assertEqual(policy.policy_id, "policy_id")
        self.assertEqual(policy.version, 1)
        self.assertTrue(policy.ip_blocking_enabled)
        self.assertEqual(policy.blocked_ips, {"1.1.1.1": True})

    def populated_blocked_ips_with_wrong_version_policy_test(self):
        policy = PatchesPolicy({
            "policy_id": "policy_id",
            "version": 2,
            "data": {
                "blocked_ips": [{
                    "ip": "1.1.1.1"
                }]
            }
        })

        self.assertEqual(policy.policy_id, "policy_id")
        self.assertEqual(policy.version, 2)
        self.assertFalse(policy.ip_blocking_enabled)
        self.assertEqual(policy.blocked_ips, {})

    def populated_blocked_ips_with_missing_ip_policy_test(self):
        policy = PatchesPolicy({
            "policy_id": "policy_id",
            "version": 1,
            "data": {
                "blocked_ips": [{
                    "ip_wrong": "1.1.1.1"
                }]
            }
        })

        self.assertEqual(policy.policy_id, "policy_id")
        self.assertEqual(policy.version, 1)
        self.assertFalse(policy.ip_blocking_enabled)
        self.assertEqual(policy.blocked_ips, {})

    def disabled_ip_blocking_is_ip_blocked_test(self):
        policy = PatchesPolicy({
            "policy_id": "policy_id",
            "version": 1,
            "data": {
            }
        })

        self.assertFalse(policy.is_ip_blocked("1.1.1.1"))

    def enabled_ip_blocking_is_ip_blocked_with_none_ip_test(self):
        policy = PatchesPolicy({
            "policy_id": "policy_id",
            "version": 1,
            "data": {
                "blocked_ips": [{
                    "ip": "1.1.1.1"
                }]
            }
        })

        self.assertFalse(policy.is_ip_blocked(None))

    def enabled_ip_blocking_is_ip_blocked_with_empty_ip_test(self):
        policy = PatchesPolicy({
            "policy_id": "policy_id",
            "version": 1,
            "data": {
                "blocked_ips": [{
                    "ip": "1.1.1.1"
                }]
            }
        })

        self.assertFalse(policy.is_ip_blocked(""))

    def enabled_ip_blocking_is_ip_blocked_with_blocked_ip_test(self):
        policy = PatchesPolicy({
            "policy_id": "policy_id",
            "version": 1,
            "data": {
                "blocked_ips": [{
                    "ip": "1.1.1.1"
                }]
            }
        })

        self.assertTrue(policy.is_ip_blocked("1.1.1.1"))

    def enabled_ip_blocking_is_ip_blocked_with_non_blocked_ip_test(self):
        policy = PatchesPolicy({
            "policy_id": "policy_id",
            "version": 1,
            "data": {
                "blocked_ips": [{
                    "ip": "1.1.1.1"
                }]
            }
        })

        self.assertFalse(policy.is_ip_blocked("2.2.2.2"))



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function

from django.test import TestCase
from mock import patch
from django.http import HttpResponse

try:
    import django
    from django.conf import settings
    settings.configure()
    settings.DATABASES = {
    'default': {
            'ENGINE': 'django.db.backends.sqlite3',
            'NAME': 'mydatabase',
    }
    }
    django.setup()
except RuntimeError:
    print("Django already setup")
    
from mock import Mock
from ....instrumentation.django.middleware.globalrequestmiddleware import GlobalRequestMiddleware
from ....instrumentation.django.middleware.afterauthmiddleware import AfterAuthMiddleware
from django.middleware.common import CommonMiddleware
from ....config import CONFIGURATION
from ....agent import TCellAgent, PolicyTypes

from httmock import urlmatch, HTTMock
import requests
import json
from django.conf import settings

from django.test.client import RequestFactory

try:
    settings.configure()
except:
    pass

settings.DEFAULT_CHARSET = 'utf-8'

class GlobalRequestMiddlewareTest(TestCase):

    def setUp(self):
        settings.ALLOWED_HOSTS = ['testserver','test.tcell.io']

        self.grm = GlobalRequestMiddleware()
        self.aam = AfterAuthMiddleware()
        self.cm = CommonMiddleware()

        rf = RequestFactory()
        self.request = rf.get('http://test.tcell.io/hello/')

        #self.request = Mock()
        self.request.session = Mock()
        self.request.session.session_key = "101012301200123"


        CONFIGURATION.version = 1
        CONFIGURATION.reverse_proxy = True
        CONFIGURATION.api_key = "Test_-ApiKey=="
        CONFIGURATION.app_id = "TestAppId-AppId"
        CONFIGURATION.host_identifier = "TestHostIdentifier"
        CONFIGURATION.uuid = "test-uuid-test-uuid"
        CONFIGURATION.fetch_policies_from_tcell = False

    def test_session_id_is_added(self):
        self.assertEqual(self.grm.process_request(self.request), None)
        self.assertEqual(self.aam.process_request(self.request), None)
        self.assertEqual(self.request._tcell_context.raw_session_id, "101012301200123")

    def test_remote_addr_is_added(self):
        self.request.META["REMOTE_ADDR"] = "1.1.2.2"
        self.grm.process_request(self.request)
        self.assertEqual(self.request._tcell_context.remote_addr, "1.1.2.2")

    def test_reverse_proxy_is_added_single(self):
        self.request.META["REMOTE_ADDR"] = "1.1.2.2"
        self.request.META["HTTP_X_FORWARDED_FOR"] = "2.3.3.2"
        self.grm.process_request(self.request)
        self.assertEqual(self.request._tcell_context.remote_addr, "2.3.3.2")

    def test_reverse_proxy_is_added_multiple(self):
        self.request.META["REMOTE_ADDR"] = "1.1.2.2"
        self.request.META["HTTP_X_FORWARDED_FOR"] = "2.2.2.2 ,3.3.3.3,4.4.4.4"
        self.grm.process_request(self.request)
        self.assertEqual(self.request._tcell_context.remote_addr, "2.2.2.2")

    def test_reverse_proxy_is_igored(self):
        CONFIGURATION.reverse_proxy = False
        self.request.META["REMOTE_ADDR"] = "1.1.2.2"
        self.request.META["HTTP_X_FORWARDED_FOR"] = "2.2.2.2,3.3.3.3,4.4.4.4"
        self.grm.process_request(self.request)
        self.assertEqual(self.request._tcell_context.remote_addr, "1.1.2.2")

    def test_reverse_proxy_with_custom_header_name(self):
        CONFIGURATION.reverse_proxy = True
        CONFIGURATION.reverse_proxy_ip_address_header = "X-ReaL-IP"
        self.request.META["REMOTE_ADDR"] = "1.1.2.2"
        self.request.META["HTTP_X_REAL_IP"] = "2.2.2.2"
        self.grm.process_request(self.request)
        self.assertEqual(self.request._tcell_context.remote_addr, "2.2.2.2")

    def test_reverse_proxy_with_custom_header_name_is_igored(self):
        CONFIGURATION.reverse_proxy = False
        CONFIGURATION.reverse_proxy_ip_address_header = "X-ReaL-IP"
        self.request.META["REMOTE_ADDR"] = "1.1.2.2"
        self.request.META["HTTP_X_REAL_IP"] = "2.2.2.2"
        self.grm.process_request(self.request)
        self.assertEqual(self.request._tcell_context.remote_addr, "1.1.2.2")

    def test_redirect_normal(self):
        def addEvent(self, event):
          event.post_process()

        TCellAgent.addEvent = addEvent
        TCellAgent.tCell_agent = TCellAgent()

        policy_json = {"http-redirect":{
          "policy_id":"nyzd", 
            "data":{
              "enabled":True,
              "whitelist":["whitelisted"],
              "block":False
            }
          }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        self.grm.process_request(self.request)
        self.aam.process_request(self.request)
        self.cm.process_request(self.request)
        self.assertEqual(self.request._tcell_context.raw_session_id, "101012301200123")
        response =  HttpResponse('hello world',content_type='application/html')
        response["Location"] = "http://www.yahoo.com/sam/5?x=asfasdfdsa"
        self.request._tcell_context.route_id = -231123123
        self.cm.process_response(self.request, response)
        self.aam.process_response(self.request, response)
        self.grm.process_response(self.request, response)

        self.assertEqual(response["location"], "http://www.yahoo.com/sam/5?x=asfasdfdsa")

    def test_redirect_blocked(self):
        def addEvent(self, event):
          event.post_process()

        TCellAgent.addEvent = addEvent
        TCellAgent.tCell_agent = TCellAgent()

        policy_json = {"http-redirect":{
          "policy_id":"nyzd", 
            "data":{
              "enabled":True,
              "whitelist":["*.google.com","*yahoo"],
              "block":True
            }
          }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        self.grm.process_request(self.request)
        self.aam.process_request(self.request)
        self.cm.process_request(self.request)
        self.assertEqual(self.request._tcell_context.raw_session_id, "101012301200123")
        response =  HttpResponse('hello world',content_type='application/html')
        response["Location"] = "http://www.yahoo.com/sam/5?x=asfasdfdsa"
        self.request._tcell_context.route_id = -231123123

        self.cm.process_response(self.request, response)
        self.aam.process_response(self.request, response)
        self.grm.process_response(self.request, response)

        self.assertEqual(response["location"], "/")

    def test_redirect_blocked_but_whitelisted(self):
        def addEvent(self, event):
          event.post_process()

        TCellAgent.addEvent = addEvent
        TCellAgent.tCell_agent = TCellAgent()

        policy_json = {"http-redirect":{
          "policy_id":"nyzd", 
            "data":{
              "enabled":True,
              "whitelist":["*.google.com","*.yahoo.com"],
              "block":True
            }
          }
        }
        TCellAgent.tCell_agent.process_policies(policy_json, cache=False)

        self.grm.process_request(self.request)
        self.aam.process_request(self.request)
        self.cm.process_request(self.request)
        self.assertEqual(self.request._tcell_context.raw_session_id, "101012301200123")
        response =  HttpResponse('hello world',content_type='application/html')
        response["Location"] = "http://www.yahoo.com/sam/5?x=asfasdfdsa"
        self.request._tcell_context.route_id = -231123123

        #self.aam.process_response(self.request, self.response)
        self.cm.process_response(self.request, response)
        self.grm.process_response(self.request, response)

        self.assertEqual(response["location"], "http://www.yahoo.com/sam/5?x=asfasdfdsa")



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

from tcell_agent.agent import TCellAgent, PolicyTypes
import os
from tcell_agent.sensor_events.httptx import HttpTxSensorEvent, FingerprintSensorEvent, LoginSuccessfulSensorEvent, LoginFailureSensorEvent
from tcell_agent.sensor_events.http_redirect import RedirectSensorEvent
import uuid
import re

from tcell_agent.sanitize import SanitizeUtils
from tcell_agent.instrumentation import handle_exception, safe_wrap_function

from future.backports.urllib.parse import urlsplit
from future.backports.urllib.parse import urlunsplit
from future.backports.urllib.parse import parse_qs

from tcell_agent.instrumentation.context import TCellInstrumentationContext

from django.http import HttpResponse
import time
import threading
import logging
logger = logging.getLogger('tcell_agent').getChild(__name__)

from django.conf import settings
from django.core.urlresolvers import get_resolver, set_urlconf
from tcell_agent.instrumentation.django.routes import get_route_table
import re
django_header_regex = re.compile('^HTTP_')

class TCellDataExposureMiddleware(object):

    def process_request(self, request):

        def form_dataex():
            dataloss_policy = TCellAgent.get_policy(PolicyTypes.DATALOSS)
            if dataloss_policy is None:
                return
            route_id = request._tcell_context.route_id
            def parameter_list(query_dict):
                for parameter in query_dict:
                    parameter_lower = parameter.lower()
                    actions = dataloss_policy.get_actions_for_form_parameter(parameter_lower, route_id)
                    if actions is None:
                        continue
                    parameter_values = query_dict.getlist(parameter,[])
                    for action in actions:
                        for parameter_value in parameter_values:
                            request._tcell_context.add_response_parameter_filter(parameter_value, action, parameter)
            parameter_list(request.GET)
            parameter_list(request.POST)

        def cookie_dataex():
            dataloss_policy = TCellAgent.get_policy(PolicyTypes.DATALOSS)
            if dataloss_policy is None:
                return
            route_id = request._tcell_context.route_id
            for cookie in request.COOKIES:
                actions = dataloss_policy.get_actions_for_cookie(cookie, route_id)
                if actions is None:
                    continue
                cookie_value = request.COOKIES.get(cookie)
                for action in actions:
                    request._tcell_context.add_response_cookie_filter(cookie_value, action, cookie)

        def header_dataex():
            dataloss_policy = TCellAgent.get_policy(PolicyTypes.DATALOSS)
            if dataloss_policy is None:
                return
            route_id = request._tcell_context.route_id
            headers = dict((django_header_regex.sub('', header), value) for (header, value) 
                        in request.META.items() if header.startswith('HTTP_'))
            for header in headers:
                actions = dataloss_policy.get_actions_for_header(header.lower(), route_id)
                if actions is None:
                    continue
                header_value = headers.get(header)
                for action in actions:
                    request._tcell_context.add_header_filter(header_value, action, header)

        safe_wrap_function("Data-Exposure for Request Forms", form_dataex)
        safe_wrap_function("Data-Exposure for Request Cookies", cookie_dataex)
        safe_wrap_function("Data-Exposure for Request Headers", header_dataex)

    def process_response(self, request, response):
        return response


# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

from django.conf import settings

from tcell_agent.agent import TCellAgent
from tcell_agent.sensor_events import AppRouteSensorEvent
from tcell_agent.instrumentation.django.utils import midVersionGreaterThanOrEqualTo

import logging
import tcell_agent.tcell_logger
import threading
LOGGER = logging.getLogger('tcell_agent').getChild(__name__)

ROUTE_TABLE = {}

def get_route_table():
    return ROUTE_TABLE

def calculate_route_id(uri):
    return str(hash(uri))

def make_route_table(send_events=True):
    def show_urls(urllist, prefix=""):
        for entry in urllist:
            try:
                pattern = entry.regex.pattern
                if pattern is None:
                    continue
                if pattern.startswith("^"):
                    pattern = pattern[1:]
                if pattern.endswith("$"):
                    pattern = pattern[:-1]
                if entry.__class__.__name__ == 'RegexURLPattern':
                    if entry.callback:
                        route_url = prefix + pattern
                        route_target = "unknown"
                        try:
                            if hasattr(entry.callback, 'func_name'):
                                route_target = entry.callback.__module__ + "." + entry.callback.func_name
                            else:
                                route_target = entry.callback.__module__ + "." + entry.callback.__name__
                        except Exception as target_exception:
                            LOGGER.debug("Exception creating target {e}".format(e=target_exception))
                        route_id = calculate_route_id(route_url)
                        ROUTE_TABLE[entry.callback] = {"pattern":route_url,
                                                       "target":route_target,
                                                       "route_id":route_id}
                if hasattr(entry, 'url_patterns'):
                    show_urls(entry.url_patterns, prefix + pattern)
            except Exception as add_route_exception:
                LOGGER.error("Exception parsing route {e}".format(e=add_route_exception))

    def send_route_table():
        try:
            for route_key in ROUTE_TABLE:
                route = ROUTE_TABLE[route_key]
                TCellAgent.discover_route(route.get("pattern"), 
                                          "*",
                                          route.get("target"),
                                          route.get("route_id")
                )
        except Exception as e:
            LOGGER.error("Exception sending route table {e}".format(e=e))

    try:
        if (midVersionGreaterThanOrEqualTo("1.9")):
            from django.core.urlresolvers import get_resolver
            resolver = get_resolver()
            show_urls(resolver.url_patterns)
        else:
            from django.core.urlresolvers import get_resolver
            resolver = get_resolver(settings.ROOT_URLCONF)
            show_urls(resolver.url_patterns)            
        send_route_table_thread = threading.Thread(target=send_route_table, args=())
        send_route_table_thread.daemon = True # Daemonize thread
        send_route_table_thread.start() 
    except Exception as e:
        LOGGER.error("Exception making the route table {e}".format(e=e))
    return get_route_table


# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

import tcell_agent
from tcell_agent.agent import TCellAgent, PolicyTypes
import os
from tcell_agent.sensor_events import HttpTxSensorEvent, FingerprintSensorEvent, LoginSuccessfulSensorEvent, LoginFailureSensorEvent
from tcell_agent.sensor_events import RedirectSensorEvent
from tcell_agent.sensor_events import ServerAgentDetailsEvent

_started = False

import logging
import tcell_agent.tcell_logger
LOGGER = logging.getLogger('tcell_agent').getChild(__name__)

def instrument_gunicorn():
    from gunicorn.arbiter import Arbiter
    old_func = Arbiter.start
    def start(self):
        if (tcell_agent.instrumentation.gunicorn_tcell._started == False):
            LOGGER.info("Staring (gunicorn) agent")
            tcell_agent.instrumentation.gunicorn_tcell._started = True
            TCellAgent.get_agent().ensure_polling_thread_running()
        return old_func(self)
    Arbiter.start = start
try:
    instrument_gunicorn()
except Exception as e:
    LOGGER.debug("Gunicorn not found")
    LOGGER.debug(e)



# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

from .. import safe_wrap_function
import re
import django

import logging
import tcell_agent.tcell_logger
LOGGER = logging.getLogger('tcell_agent').getChild(__name__)

isDjango15 = False
django15or16 = False
try:
  isDjango15 = django.get_version().startswith('1.5')
except:
  LOGGER.warn("Could not determine Django version for compatibility tests")

def midVersionGreaterThanOrEqualTo(version_string):
  try:
    django_midv = django.get_version().split(".")[:2]
    comparison_midv = version_string.split(".")[:2]
    if (int(django_midv[0]) >= int(comparison_midv[0]) and int(django_midv[1]) >= int(comparison_midv[1])):
      return True
  except:
    LOGGER.warn("Could not determine Django midversion for compatibility tests")
  return False

try:
  django15or16 = isDjango15 or django.get_version().startswith('1.6')
except:
  LOGGER.warn("Could not determine Django version for compatibility tests")


regex_http_          = re.compile(r'^HTTP_.+$')
regex_content_type   = re.compile(r'^CONTENT_TYPE$')
regex_content_length = re.compile(r'^CONTENT_LENGTH$')

def header_keys_from_django_request(request):
    regex = re.compile('^HTTP_')
    headers = list(regex.sub('', header) for (header, value) 
                in request.META.items() if (header.startswith('HTTP_') or header == "CONTENT_TYPE" or header == "CONTENT_LENGTH"))
    return headers




from tcell_agent.policies.appsensor.sensor import sendEvent
import traceback

class MiscSensor(object):

    def __init__(self, policy_json=None):
        self.csrf_exception_enabled = False
        self.sql_exception_enabled = False
        self.excluded_route_ids = {}

        if policy_json is not None:
            self.csrf_exception_enabled = policy_json.get("csrf_exception_enabled", False)
            self.sql_exception_enabled = policy_json.get("sql_exception_enabled", False)

            for route_id in policy_json.get("exclude_routes", []):
                self.excluded_route_ids[route_id] = True

    def csrf_rejected(self, appsensor_meta, reason):
        if not self.csrf_exception_enabled:
            return

        if self.excluded_route_ids.get(appsensor_meta.route_id, False):
            return False

        sendEvent(
          appsensor_meta,
          "excsrf",
          None,
          None)

    def sql_exception_detected(self, database, appsensor_meta, exc_type, exc_value, tb):
        if not self.sql_exception_enabled:
            return

        if self.excluded_route_ids.get(appsensor_meta.route_id, False):
            return False

        dj_exc_type = getattr(database, "ProgrammingError")
        if not issubclass(exc_type, dj_exc_type):
            return

        sendEvent(
          appsensor_meta,
          "exsql",
          exc_type.__name__,
          None,
          payload=traceback.print_tb(tb))

    def __str__(self):
        return "<%s csrf_exception_enabled: %s sql_exception_enabled: %s>" % \
            (type(self).__name__, self.csrf_exception_enabled, self.sql_exception_enabled)



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from __future__ import print_function

import unittest
from ....policies.http_redirect_policy import HttpRedirectPolicy

class HttpRedirectPolicyTest(unittest.TestCase):
    def min_header_test(self):
        policy_json = {"policy_id":"xyzd"}
        policy = HttpRedirectPolicy()
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.policy_id, "xyzd")
        self.assertEqual(policy.enabled, False)
        self.assertEqual(policy.block, False)
        self.assertEqual(policy.whitelist, [])

    def small_header_test(self):
        policy_json = {"policy_id":"nyzd", "data":{"enabled":True}}
        policy = HttpRedirectPolicy()
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.policy_id, "nyzd")
        self.assertEqual(policy.enabled, True)
        self.assertEqual(policy.block, False)
        self.assertEqual(policy.whitelist, [])

    def large_header_test(self):
        policy_json = {"policy_id":"nyzd",
                       "data":
                           {
                               "enabled":True,
                               "whitelist":["whitelisted"],
                               "block":True
                           }
                      }
        policy = HttpRedirectPolicy()
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.policy_id, "nyzd")
        self.assertEqual(policy.enabled, True)
        self.assertEqual(policy.block, True)
        whitelist = ["whitelisted"]
        compiled_re = [policy.wildcard_re(item) for item in whitelist]
        self.assertEqual(policy.whitelist, compiled_re)

    def same_domain_redirect_test(self):
        policy_json = {"policy_id":"nyzd",
                       "data":
                           {
                               "enabled":True,
                               "whitelist":["whitelisted"],
                               "block":True
                           }
                      }
        policy = HttpRedirectPolicy()
        policy.loadFromJson(policy_json)
        self.assertEqual(policy.policy_id, "nyzd")
        self.assertEqual(policy.enabled, True)
        self.assertEqual(policy.block, True)
        whitelist = ["whitelisted"]
        compiled_re = [policy.wildcard_re(item) for item in whitelist]
        self.assertEqual(policy.whitelist, compiled_re)

        check = policy.process_location(
            "0.1.1.0",
            "GET",
            "localhost:8011",
            "/etc/123",
            200,
            "http://localhost:8011/abc/def")

        self.assertEqual(check, "http://localhost:8011/abc/def")



# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function
from . import SensorEvent
from tcell_agent.sanitize import SanitizeUtils
from future.backports.urllib.parse import urlsplit

class ServerAgentPackagesEvent(SensorEvent):
    
    def __init__(self):
        super(ServerAgentPackagesEvent, self).__init__("server_agent_packages", ensure_delivery=True, flush_right_away=True)
        self.flush = True
        self.ensure = True
        self["packages"] = []
    
    def add_package(self, name, version, license=None):
        package = {
          "n":name, 
          "v":version,
          }
        if license is not None:
          package["l"] = license
        self["packages"].append(package)


# -*- coding: utf-8 -*-
# Copyright (C) 2015 tCell.io, Inc. - All Rights Reserved

from __future__ import unicode_literals
from __future__ import print_function

class InstrumentedMethod(object):
    def __init__(self, obj, func_name, target):
        self.obj = obj
        self.func_name = func_name
        self.target = target

class InstrumentationManager(object):
    _instrumented = {}
    initial_instrumentation_run = False

    @classmethod
    def instrument(cls, original_object, original_method_str, new_method):
        target_method = getattr(original_object, original_method_str)
        already_wrapped = hasattr(target_method, "__tcell_instrumentation__original_method__")
        if already_wrapped:
            return
        def wrapped_func(*args, **kwargs):
            return new_method(target_method, *args, **kwargs)
        wrapped_func.__tcell_instrumentation__original_method__ = target_method
        wrapped_func.__tcell_instrumentation__original_object__ = original_object
        wrapped_func.__tcell_instrumentation__original_method_str__ = original_method_str
        setattr(original_object, original_method_str, wrapped_func)
        return

    @classmethod
    def is_instrumented(cls, target_method):
        already_wrapped = hasattr(target_method, "__tcell_instrumentation__original_method__")
        return already_wrapped

    @classmethod
    def remove_instrumentation(cls, original_object, original_func_name, target_method):
        pass
